#!/usr/bin/env python3
from __future__ import annotations

from plumbum import colors

with colors.fg.red:
    print("This is in red")

print("This is completely restored, even if an exception is thrown!")

print("The library will restore color on exiting automatically.")
print(colors.bold["This is bold and exciting!"])
print(colors.bg.cyan | "This is on a cyan background.")
print(colors.fg[42] | "If your terminal supports 256 colors, this is colorful!")
print()
for c in colors:
    print(c + "\u2588", end="")
colors.reset()
print()
print("Colors can be reset " + colors.underline["Too!"])
for c in colors[:16]:
    print(c["This is in color!"])

colors.red()
print("This should clean up the color automatically on program exit...")



#!/usr/bin/env python3
"""
Examples::

    $ python3 geet.py
    no command given

    $ python3 geet.py leet
    unknown command 'leet'

    $ python3 geet.py --help
    geet v1.7.2
    The l33t version control

    Usage: geet.py [SWITCHES] [SUBCOMMAND [SWITCHES]] args...
    Meta-switches:
        -h, --help                 Prints this help message and quits
        -v, --version              Prints the program's version and quits

    Subcommands:
        commit                     creates a new commit in the current branch; see
                                   'geet commit --help' for more info
        push                       pushes the current local branch to the remote
                                   one; see 'geet push --help' for more info

    $ python3 geet.py commit --help
    geet commit v1.7.2
    creates a new commit in the current branch

    Usage: geet commit [SWITCHES]
    Meta-switches:
        -h, --help                 Prints this help message and quits
        -v, --version              Prints the program's version and quits

    Switches:
        -a                         automatically add changed files
        -m VALUE:str               sets the commit message; required

    $ python3 geet.py commit -m "foo"
    committing...
"""

from __future__ import annotations

try:
    import colorama

    colorama.init()
except ImportError:
    pass

from plumbum import cli, colors


class Geet(cli.Application):
    SUBCOMMAND_HELPMSG = False
    DESCRIPTION = colors.yellow | """The l33t version control"""
    PROGNAME = colors.green
    VERSION = colors.blue | "1.7.2"
    COLOR_USAGE_TITLE = colors.bold | colors.magenta
    COLOR_USAGE = colors.magenta

    _group_names = ["Meta-switches", "Switches", "Sub-commands"]

    COLOR_GROUPS = dict(
        zip(_group_names, [colors.do_nothing, colors.skyblue1, colors.yellow])
    )

    COLOR_GROUP_TITLES = dict(
        zip(
            _group_names,
            [colors.bold, colors.bold | colors.skyblue1, colors.bold | colors.yellow],
        )
    )

    verbosity = cli.SwitchAttr(
        "--verbosity",
        cli.Set("low", "high", "some-very-long-name", "to-test-wrap-around"),
        help=colors.cyan
        | "sets the verbosity level of the geet tool. doesn't really do anything except for testing line-wrapping "
        "in help " * 3,
    )
    verbositie = cli.SwitchAttr(
        "--verbositie",
        cli.Set("low", "high", "some-very-long-name", "to-test-wrap-around"),
        help=colors.hotpink
        | "sets the verbosity level of the geet tool. doesn't really do anything except for testing line-wrapping "
        "in help " * 3,
    )


@Geet.subcommand(colors.red | "commit")
class GeetCommit(cli.Application):
    """creates a new commit in the current branch"""

    auto_add = cli.Flag("-a", help="automatically add changed files")
    message = cli.SwitchAttr("-m", str, mandatory=True, help="sets the commit message")

    def main(self):
        print("committing...")


GeetCommit.unbind_switches("-v", "--version")


@Geet.subcommand("push")
class GeetPush(cli.Application):
    """pushes the current local branch to the remote one"""

    tags = cli.Flag("--tags", help="whether to push tags (default is False)")

    def main(self, remote, branch="master"):
        print(f"pushing to {remote}/{branch}...")


if __name__ == "__main__":
    Geet.run()



#!/usr/bin/env python3
from __future__ import annotations

from plumbum import FG, cli, local
from plumbum.cmd import convert, pdflatex
from plumbum.path.utils import delete


def image_comp(item):
    pdflatex["-shell-escape", item] & FG
    print("Converting", item)
    convert[item.with_suffix(".svg"), item.with_suffix(".png")] & FG

    delete(
        item.with_suffix(".log"),
        item.with_suffix(".aux"),
    )


class MyApp(cli.Application):
    def main(self, *srcfiles):
        print("Tex files should start with:")
        print(r"\documentclass[tikz,convert={outfile=\jobname.svg}]{standalone}")
        items = map(cli.ExistingFile, srcfiles) if srcfiles else local.cwd // "*.tex"
        for item in items:
            image_comp(item)


if __name__ == "__main__":
    MyApp.run()



from __future__ import annotations

from plumbum.machines.local import LocalCommand, LocalMachine, local
from plumbum.machines.remote import BaseRemoteMachine, RemoteCommand
from plumbum.machines.ssh_machine import PuttyMachine, SshMachine

__all__ = (
    "BaseRemoteMachine",
    "LocalCommand",
    "LocalMachine",
    "PuttyMachine",
    "RemoteCommand",
    "SshMachine",
    "local",
)



r"""
Plumbum Shell Combinators
-------------------------
A wrist-handy library for writing shell-like scripts in Python, that can serve
as a ``Popen`` replacement, and much more::

    >>> from plumbum.cmd import ls, grep, wc, cat
    >>> ls()
    'build.py\ndist\ndocs\nLICENSE\nplumbum\nREADME.rst\nsetup.py\ntests\ntodo.txt\n'
    >>> chain = ls["-a"] | grep["-v", "py"] | wc["-l"]
    >>> print(chain)
    /bin/ls -a | /bin/grep -v py | /usr/bin/wc -l
    >>> chain()
    '12\n'
    >>> ((ls["-a"] | grep["-v", "py"]) > "/tmp/foo.txt")()
    ''
    >>> ((cat < "/tmp/foo.txt") | wc["-l"])()
    '12\n'
    >>> from plumbum import local, FG, BG
    >>> with local.cwd("/tmp"):
    ...     (ls | wc["-l"]) & FG
    ...
    13              # printed directly to the interpreter's stdout
    >>> (ls | wc["-l"]) & BG
    <Future ['/usr/bin/wc', '-l'] (running)>
    >>> f = _
    >>> f.stdout    # will wait for the process to terminate
    '9\n'

Plumbum includes local/remote path abstraction, working directory and environment
manipulation, process execution, remote process execution over SSH, tunneling,
SCP-based upload/download, and a {arg|opt}parse replacement for the easy creation
of command-line interface (CLI) programs.

See https://plumbum.readthedocs.io for full details
"""

from __future__ import annotations

# Avoids a circular import error later
import plumbum.path  # noqa: F401
from plumbum.commands import (
    BG,
    ERROUT,
    FG,
    NOHUP,
    RETCODE,
    TEE,
    TF,
    CommandNotFound,
    ProcessExecutionError,
    ProcessLineTimedOut,
    ProcessTimedOut,
)
from plumbum.machines import BaseRemoteMachine, PuttyMachine, SshMachine, local
from plumbum.path import LocalPath, Path, RemotePath
from plumbum.version import version

__author__ = "Tomer Filiba (tomerfiliba@gmail.com)"
__version__ = version

__all__ = (
    "BG",
    "ERROUT",
    "FG",
    "NOHUP",
    "RETCODE",
    "TEE",
    "TF",
    "BaseRemoteMachine",
    "CommandNotFound",
    "LocalPath",
    "Path",
    "ProcessExecutionError",
    "ProcessLineTimedOut",
    "ProcessTimedOut",
    "PuttyMachine",
    "RemotePath",
    "SshMachine",
    "__author__",
    "__version__",
    "cmd",
    "local",
)

from . import cmd


def __dir__():
    "Support nice tab completion"
    return __all__



#!/usr/bin/env python3
from __future__ import annotations

import pytest

from plumbum import colors
from plumbum.colorlib import htmlcolors
from plumbum.colorlib.styles import ANSIStyle as Style
from plumbum.colorlib.styles import ColorNotFound


class TestImportColors:
    def testDifferentImports(self):
        from plumbum.colors import bold
        from plumbum.colors.fg import red

        assert str(red) == str(colors.red)
        assert str(bold) == str(colors.bold)


class TestANSIColor:
    def setup_method(self, method):  # noqa: ARG002
        colors.use_color = True

    def testColorSlice(self):
        vals = colors[:8]
        assert len(vals) == 8
        assert vals[1] == colors.red
        vals = colors[40:50]
        assert len(vals) == 10
        assert vals[1] == colors.full(41)

    def testLoadNumericalColor(self):
        assert colors.full(2) == colors[2]
        assert colors.simple(2) == colors(2)
        assert colors(54) == colors[54]
        assert colors(1, 30, 77) == colors.rgb(1, 30, 77)
        assert colors[1, 30, 77] == colors.rgb(1, 30, 77)

    def testColorStrings(self):
        assert colors.reset == "\033[0m"
        assert colors.bold == "\033[1m"
        assert colors.fg.reset == "\033[39m"

    def testNegateIsReset(self):
        assert colors.reset == ~colors
        assert colors.fg.reset == ~colors.fg
        assert colors.bg.reset == ~colors.bg

    def testFromPreviousColor(self):
        assert colors(colors.red) == colors.red
        assert colors(colors.bg.red) == colors.bg.red
        assert colors(colors.bold) == colors.bold

    def testFromCode(self):
        assert colors("\033[31m") == colors.red

    def testEmptyStyle(self):
        assert str(colors()) == ""
        assert str(colors("")) == ""
        assert str(colors(None)) == ""

    def testLoadColorByName(self):
        assert colors["LightBlue"] == colors.fg["LightBlue"]
        assert colors.bg["light_green"] == colors.bg["LightGreen"]
        assert colors["DeepSkyBlue1"] == colors["#00afff"]
        assert colors["DeepSkyBlue1"] == colors.hex("#00afff")

        assert colors["DeepSkyBlue1"] == colors[39]
        assert colors.DeepSkyBlue1 == colors[39]
        assert colors.deepskyblue1 == colors[39]
        assert colors.Deep_Sky_Blue1 == colors[39]
        assert colors.red == colors.RED

        with pytest.raises(AttributeError):
            colors.Notacolorsatall  # noqa: B018

    def testMultiColor(self):
        sumcolors = colors.bold & colors.blue
        assert colors.bold.reset & colors.fg.reset == ~sumcolors

    def testSums(self):
        # Sums should not be communitave, last one is used
        assert colors.red == colors.blue & colors.red
        assert colors.bg.green == colors.bg.red & colors.bg.green

    def testRepresentations(self):
        colors1 = colors.full(87)
        assert colors1 == colors.DarkSlateGray2
        assert colors1.basic == colors.DarkSlateGray2
        assert str(colors1.basic) == str(colors.LightGray)

        colors2 = colors.rgb(1, 45, 214)
        assert str(colors2.full) == str(colors.Blue3A)

    def testFromAnsi(self):
        for c in colors[1:7]:
            assert c == colors.from_ansi(str(c))
        for c in colors.bg[1:7]:
            assert c == colors.from_ansi(str(c))
        for c in colors:
            assert c == colors.from_ansi(str(c))
        for c in colors.bg:
            assert c == colors.from_ansi(str(c))
        for c in colors[:16]:
            assert c == colors.from_ansi(str(c))
        for c in colors.bg[:16]:
            assert c == colors.from_ansi(str(c))
        for c in (colors.bold, colors.underline, colors.italics):
            assert c == colors.from_ansi(str(c))

        col = colors.bold & colors.fg.green & colors.bg.blue & colors.underline
        assert col == colors.from_ansi(str(col))
        col = colors.reset
        assert col == colors.from_ansi(str(col))

    def testWrappedColor(self):
        string = "This is a string"
        wrapped = "\033[31mThis is a string\033[39m"
        assert colors.red.wrap(string) == wrapped
        assert colors.red | string == wrapped
        assert colors.red[string] == wrapped

        newcolors = colors.blue & colors.underline
        assert newcolors[string] == string | newcolors
        assert newcolors.wrap(string) == string | colors.blue & colors.underline

    def testUndoColor(self):
        assert ~colors.fg == "\033[39m"
        assert ~colors.bg == "\033[49m"
        assert ~colors.bold == "\033[22m"
        assert ~colors.dim == "\033[22m"
        for i in range(7):
            assert ~colors(i) == "\033[39m"
            assert ~colors.bg(i) == "\033[49m"
            assert ~colors.fg(i) == "\033[39m"
            assert ~colors.bg(i) == "\033[49m"
        for i in range(256):
            assert ~colors.fg[i] == "\033[39m"
            assert ~colors.bg[i] == "\033[49m"
        assert ~colors.reset == "\033[0m"
        assert colors.do_nothing == ~colors.do_nothing

        assert colors.bold.reset == ~colors.bold

    def testLackOfColor(self):
        Style.use_color = False
        assert colors.fg.red == ""
        assert ~colors.fg == ""
        assert colors.fg["LightBlue"] == ""

    def testFromHex(self):
        with pytest.raises(ColorNotFound):
            colors.hex("asdf")

        with pytest.raises(ColorNotFound):
            colors.hex("#1234Z2")

        with pytest.raises(ColorNotFound):
            colors.hex(12)

    def testDirectCall(self, capsys):
        colors.blue()
        assert capsys.readouterr()[0] == str(colors.blue)

    def testPrint(self, capsys):
        colors.yellow.print("This is printed to stdout", end="")
        assert capsys.readouterr()[0] == str(
            colors.yellow.wrap("This is printed to stdout")
        )


class TestHTMLColor:
    def test_html(self):
        red_tagged = '<font color="#C00000">This is tagged</font>'
        assert htmlcolors.red["This is tagged"] == red_tagged
        assert "This is tagged" | htmlcolors.red == red_tagged

        twin_tagged = '<font color="#C00000"><em>This is tagged</em></font>'
        assert "This is tagged" | htmlcolors.red & htmlcolors.em == twin_tagged
        assert "This is tagged" | htmlcolors.em & htmlcolors.red == twin_tagged
        assert htmlcolors.em & htmlcolors.red | "This is tagged" == twin_tagged



from __future__ import annotations

from plumbum.cli import Application, ExistingFile


class App(Application):
    def main(self, file: ExistingFile):
        print(f"file={file.name}")


def test_access_annotations(capsys):
    _, rc = App.run(["prog", "pyproject.toml"], exit=False)
    assert rc == 0
    stdout, _ = capsys.readouterr()
    assert "file=pyproject.toml" in stdout



"""
Terminal-related utilities
--------------------------
"""

from __future__ import annotations

import contextlib
import os
import sys

from plumbum import local

from .progress import Progress
from .termsize import get_terminal_size

__all__ = [
    "Progress",
    "ask",
    "choose",
    "get_terminal_size",
    "get_terminal_size",
    "prompt",
    "readline",
]


def __dir__() -> list[str]:
    return __all__


def readline(message: str = "") -> str:
    """Gets a line of input from the user (stdin)"""
    sys.stdout.write(message)
    sys.stdout.flush()
    return sys.stdin.readline()


def ask(question: str, default: bool | None = None) -> bool:
    """
    Presents the user with a yes/no question.

    :param question: The question to ask
    :param default: If ``None``, the user must answer. If ``True`` or ``False``, lack of response is
                    interpreted as the default option

    :returns: the user's choice
    """
    question = question.rstrip().rstrip("?").rstrip() + "?"
    if default is None:
        question += " (y/n) "
    elif default:
        question += " [Y/n] "
    else:
        question += " [y/N] "

    while True:
        try:
            answer = readline(question).strip().lower()
        except EOFError:
            answer = None
        if answer in {"y", "yes"}:
            return True
        if answer in {"n", "no"}:
            return False
        if not answer and default is not None:
            return default
        sys.stdout.write("Invalid response, please try again\n")


def choose(question, options, default=None):
    """Prompts the user with a question and a set of options, from which the user needs to choose.

    :param question: The question to ask
    :param options: A set of options. It can be a list (of strings or two-tuples, mapping text
                    to returned-object) or a dict (mapping text to returned-object).``
    :param default: If ``None``, the user must answer. Otherwise, lack of response is interpreted
                    as this answer

    :returns: The user's choice

    Example::

        ans = choose("What is your favorite color?", ["blue", "yellow", "green"], default = "yellow")
        # `ans` will be one of "blue", "yellow" or "green"

        ans = choose("What is your favorite color?",
                {"blue" : 0x0000ff, "yellow" : 0xffff00 , "green" : 0x00ff00}, default = 0x00ff00)
        # this will display "blue", "yellow" and "green" but return a numerical value
    """
    if hasattr(options, "items"):
        options = options.items()
    sys.stdout.write(question.rstrip() + "\n")
    choices = {}
    defindex = None
    for i, item in enumerate(options, 1):
        if isinstance(item, (tuple, list)) and len(item) == 2:
            text = item[0]
            val = item[1]
        else:
            text = item
            val = item
        choices[i] = val
        if default is not None and default == val:
            defindex = i
        sys.stdout.write(f"({i}) {text}\n")
    if default is not None:
        msg = f"Choice [{default}]: " if defindex is None else f"Choice [{defindex}]: "
    else:
        msg = "Choice: "
    while True:
        try:
            choice = readline(msg).strip()
        except EOFError:
            choice = ""
        if not choice and default:
            return default
        try:
            choice = int(choice)
            if choice not in choices:
                raise ValueError()
        except ValueError:
            sys.stdout.write("Invalid choice, please try again\n")
            continue
        return choices[choice]


def prompt(
    question,
    type=str,  # pylint: disable=redefined-builtin
    default=NotImplemented,
    validator=lambda _: True,
):
    """
    Presents the user with a validated question, keeps asking if validation does not pass.

    :param question: The question to ask
    :param type: The type of the answer, defaults to str
    :param default: The default choice
    :param validator: An extra validator called after type conversion, can raise ValueError or return False to trigger a retry.

    :returns: the user's choice
    """
    question = question.rstrip(" \t:")
    if default is not NotImplemented:
        question += f" [{default}]"
    question += ": "
    while True:
        try:
            ans = readline(question).strip()
        except EOFError:
            ans = ""

        if not ans:
            if default is not NotImplemented:
                # sys.stdout.write("\b%s\n" % (default,))
                return default
            continue
        try:
            ans = type(ans)
        except (TypeError, ValueError) as ex:
            sys.stdout.write(f"Invalid value ({ex}), please try again\n")
            continue

        try:
            valid = validator(ans)
        except ValueError as ex:
            sys.stdout.write(f"{ex}, please try again\n")
            continue

        if not valid:
            sys.stdout.write("Value not in specified range, please try again\n")
            continue
        return ans


def hexdump(data_or_stream, bytes_per_line=16, aggregate=True):
    """Convert the given bytes (or a stream with a buffering ``read()`` method) to hexdump-formatted lines,
    with possible aggregation of identical lines. Returns a generator of formatted lines.
    """
    if hasattr(data_or_stream, "read"):

        def read_chunk():
            while True:
                buf = data_or_stream.read(bytes_per_line)
                if not buf:
                    break
                yield buf

    else:

        def read_chunk():
            for i in range(0, len(data_or_stream), bytes_per_line):
                yield data_or_stream[i : i + bytes_per_line]

    prev = None
    skipped = False
    for i, chunk in enumerate(read_chunk()):
        hexd = " ".join(f"{ord(ch):02x}" for ch in chunk)
        text = "".join(ch if 32 <= ord(ch) < 127 else "." for ch in chunk)
        if aggregate and prev == chunk:
            skipped = True
            continue
        prev = chunk
        if skipped:
            yield "*"
        hexd_ljust = hexd.ljust(bytes_per_line * 3, " ")
        yield f"{i * bytes_per_line:06x} | {hexd_ljust}| {text}"
        skipped = False


def pager(rows, pagercmd=None):  # pragma: no cover
    """Opens a pager (e.g., ``less``) to display the given text. Requires a terminal.

    :param rows: a ``bytes`` or a list/iterator of "rows" (``bytes``)
    :param pagercmd: the pager program to run. Defaults to ``less -RSin``
    """
    if not pagercmd:
        pagercmd = local["less"]["-RSin"]
    if hasattr(rows, "splitlines"):
        rows = rows.splitlines()

    pg = pagercmd.popen(stdout=None, stderr=None)
    try:
        for row in rows:
            line = f"{row}\n"
            try:
                pg.stdin.write(line)
                pg.stdin.flush()
            except OSError:
                break
        pg.stdin.close()
        pg.wait()
    finally:
        with contextlib.suppress(Exception):
            rows.close()
        if pg and pg.poll() is None:
            with contextlib.suppress(Exception):
                pg.terminate()
            os.system("reset")



# Setting French as system language
from __future__ import annotations

import importlib
import locale

import pytest

import plumbum.cli
import plumbum.cli.i18n


def reload_cli():
    importlib.reload(plumbum.cli.i18n)
    importlib.reload(plumbum.cli.switches)
    importlib.reload(plumbum.cli.application)
    importlib.reload(plumbum.cli)


@pytest.fixture
def french():
    try:
        locale.setlocale(locale.LC_ALL, "fr_FR.utf-8")
        reload_cli()
        yield
    except locale.Error:
        pytest.skip(
            "No fr_FR locale found, run 'sudo locale-gen fr_FR.UTF-8' to run this test"
        )
    finally:
        locale.setlocale(locale.LC_ALL, "")
        reload_cli()


@pytest.mark.usefixtures("french")
def test_nolang_switches():
    class Simple(plumbum.cli.Application):
        foo = plumbum.cli.SwitchAttr("--foo")

        def main(self):
            pass

    _, rc = Simple.run(["foo", "-h"], exit=False)
    assert rc == 0
    _, rc = Simple.run(["foo", "--version"], exit=False)
    assert rc == 0


@pytest.mark.usefixtures("french")
def test_help_lang(capsys):
    class Simple(plumbum.cli.Application):
        foo = plumbum.cli.SwitchAttr("--foo")

        def main(self):
            pass

    _, rc = Simple.run(["foo", "-h"], exit=False)
    assert rc == 0
    stdout, _ = capsys.readouterr()
    assert "Utilisation" in stdout
    assert "Imprime ce message d'aide et sort" in stdout



from __future__ import annotations

import itertools
import logging
import os
import re
import tempfile

import pytest

SDIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def testdir():
    os.chdir(SDIR)


@pytest.fixture
def cleandir():
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)


# Pulled from https://github.com/reece/pytest-optional-tests

"""implements declaration of optional tests using pytest markers

The MIT License (MIT)

Copyright (c) 2019 Reece Hart

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""


_logger = logging.getLogger(__name__)

marker_re = re.compile(r"^\s*(?P<marker>\w+)(:\s*(?P<description>.*))?")


def pytest_addoption(parser):
    group = parser.getgroup("collect")
    group.addoption(
        "--run-optional-tests",
        action="append",
        dest="run_optional_tests",
        default=None,
        help="Optional test markers to run, multiple and/or comma separated okay",
    )
    parser.addini(
        "optional_tests", "list of optional markers", type="linelist", default=""
    )


def pytest_configure(config):
    # register all optional tests declared in ini file as markers
    # https://docs.pytest.org/en/latest/writing_plugins.html#registering-custom-markers
    ot_ini = config.inicfg.get("optional_tests").strip().splitlines()
    for ot_ in ot_ini:
        # ot should be a line like "optmarker: this is an opt marker", as with markers section
        config.addinivalue_line("markers", ot_)
    ot_markers = {marker_re.match(ln).group(1) for ln in ot_ini}

    # collect requested optional tests
    ot_run = config.getoption("run_optional_tests")
    if ot_run:
        ot_run = list(itertools.chain.from_iterable(a.split(",") for a in ot_run))
    else:
        ot_run = config.inicfg.get("run_optional_tests", [])
        if ot_run:
            ot_run = list(re.split(r"[,\s]+", ot_run))
    ot_run = set(ot_run)

    _logger.info("optional tests to run: %s", ot_run)
    if ot_run:
        unknown_tests = ot_run - ot_markers
        if unknown_tests:
            raise ValueError(
                "Requested execution of undeclared optional tests: {}".format(
                    ", ".join(unknown_tests)
                )
            )

    config._ot_markers = set(ot_markers)
    config._ot_run = set(ot_run)


def pytest_collection_modifyitems(config, items):
    # https://stackoverflow.com/a/50114028/342839
    ot_markers = config._ot_markers
    ot_run = config._ot_run

    skips = {}
    for item in items:
        marker_names = {m.name for m in item.iter_markers()}
        if not marker_names:
            continue
        test_otms = marker_names & ot_markers
        if not test_otms:
            # test is not marked with any optional marker
            continue
        if test_otms & ot_run:
            # test is marked with an enabled optional test; don't skip
            continue
        mns = str(marker_names)
        if mns not in skips:
            skips[mns] = pytest.mark.skip(
                reason="Skipping; marked with disabled optional tests ({})".format(
                    ", ".join(marker_names)
                )
            )
        item.add_marker(skips[mns])



from __future__ import annotations

import unittest

from parallel import Cluster

from plumbum import SshMachine, local

TEST_HOST = "127.0.0.1"


class TestParallel(unittest.TestCase):
    def setUp(self):
        self.remotes = []

    def connect(self):
        m = SshMachine(TEST_HOST)
        self.remotes.append(m)
        return m

    def tearDown(self):
        for m in self.remotes:
            m.close()

    def test_parallel(self):
        m = Cluster(local, local)
        import time

        t = time.time()
        ret = m["sleep"]("2")
        assert len(ret) == 2
        assert 2 <= time.time() - t < 4

    def test_locals(self):
        m = Cluster(local, local, local)
        # we should get 3 different proc ids
        ret = m["bash"]["-c"]["echo $$"]()
        ret = list(map(int, ret))
        assert len(set(ret)) == 3

    def test_sessions(self):
        m = Cluster(local, self.connect(), local, self.connect())
        # we should get 4 different proc ids
        ret, stdout, _ = m.session().run("echo $$")
        ret = [int(pid) for pid in stdout]
        assert len(set(ret)) == 4

    def test_commands(self):
        cmds = local["echo"]["1"] & local["echo"]["2"]
        ret = cmds()
        a, b = map(int, ret)
        assert (a, b) == (1, 2)



from __future__ import annotations

import contextlib
import errno
import os
import signal
import subprocess
import sys
import time
import traceback

from plumbum.commands.processes import ProcessExecutionError


class _fake_lock:
    """Needed to allow normal os.exit() to work without error"""

    @staticmethod
    def acquire(_):
        return True

    @staticmethod
    def release():
        pass


def posix_daemonize(command, cwd, stdout=None, stderr=None, append=True):
    if stdout is None:
        stdout = os.devnull
    if stderr is None:
        stderr = stdout

    MAX_SIZE = 16384
    rfd, wfd = os.pipe()
    argv = command.formulate()
    firstpid = os.fork()
    if firstpid == 0:
        # first child: become session leader
        os.close(rfd)
        rc = 0
        try:
            os.setsid()
            os.umask(0)
            stdin = open(os.devnull, encoding="utf-8")
            stdout = open(stdout, "a" if append else "w", encoding="utf-8")
            stderr = open(stderr, "a" if append else "w", encoding="utf-8")
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            proc = command.popen(
                cwd=cwd,
                close_fds=True,
                stdin=stdin.fileno(),
                stdout=stdout.fileno(),
                stderr=stderr.fileno(),
            )
            os.write(wfd, str(proc.pid).encode("utf8"))
        except Exception:
            rc = 1
            tbtext = "".join(traceback.format_exception(*sys.exc_info()))[-MAX_SIZE:]
            os.write(wfd, tbtext.encode("utf8"))
        finally:
            os.close(wfd)
            os._exit(rc)

    # wait for first child to die
    os.close(wfd)
    _, rc = os.waitpid(firstpid, 0)
    output = os.read(rfd, MAX_SIZE)
    os.close(rfd)
    with contextlib.suppress(UnicodeError):
        output = output.decode("utf8")
    if rc == 0 and output.isdigit():
        secondpid = int(output)
    else:
        raise ProcessExecutionError(argv, rc, "", output)
    proc = subprocess.Popen.__new__(subprocess.Popen)
    proc._child_created = True
    proc.returncode = None
    proc.stdout = None
    proc.stdin = None
    proc.stderr = None
    proc.pid = secondpid
    proc.universal_newlines = False
    proc._input = None
    proc._waitpid_lock = _fake_lock()
    proc._communication_started = False
    proc.args = argv
    proc.argv = argv

    def poll(self=proc):
        if self.returncode is None:
            try:
                os.kill(self.pid, 0)
            except OSError as ex:
                if ex.errno == errno.ESRCH:
                    # process does not exist
                    self.returncode = 0
                else:
                    raise
        return self.returncode

    def wait(self=proc):
        while self.returncode is None:
            if self.poll() is None:
                time.sleep(0.5)
        return proc.returncode

    proc.poll = poll
    proc.wait = wait
    return proc


def win32_daemonize(command, cwd, stdout=None, stderr=None, append=True):
    if stdout is None:
        stdout = os.devnull
    if stderr is None:
        stderr = stdout
    DETACHED_PROCESS = 0x00000008
    stdin = open(os.devnull, encoding="utf-8")
    stdout = open(stdout, "a" if append else "w", encoding="utf-8")
    stderr = open(stderr, "a" if append else "w", encoding="utf-8")
    return command.popen(
        cwd=cwd,
        stdin=stdin.fileno(),
        stdout=stdout.fileno(),
        stderr=stderr.fileno(),
        creationflags=subprocess.CREATE_NEW_PROCESS_GROUP | DETACHED_PROCESS,
    )



#!/usr/bin/env python3
"""
$ python3 simple_cli.py --help
simple_cli.py v1.0

Usage: simple_cli.py [SWITCHES] srcfiles...
Meta-switches:
    -h, --help                 Prints this help message and quits
    --version                  Prints the program's version and quits

Switches:
    -I VALUE:str               Specify include directories; may be given
                               multiple times
    --loglevel LEVEL:int       Sets the log-level of the logger
    -v, --verbose              Enable verbose mode

$ python3 simple_cli.py x.cpp y.cpp z.cpp
Verbose: False
Include dirs: []
Compiling: ('x.cpp', 'y.cpp', 'z.cpp')

$ python3 simple_cli.py -v
Verbose: True
Include dirs: []
Compiling: ()

$ python3 simple_cli.py -v -Ifoo/bar -Ispam/eggs
Verbose: True
Include dirs: ['foo/bar', 'spam/eggs']
Compiling: ()

$ python3 simple_cli.py -v -I foo/bar -Ispam/eggs x.cpp y.cpp z.cpp
Verbose: True
Include dirs: ['foo/bar', 'spam/eggs']
Compiling: ('x.cpp', 'y.cpp', 'z.cpp')
"""

from __future__ import annotations

import logging

from plumbum import cli


class MyCompiler(cli.Application):
    verbose = cli.Flag(["-v", "--verbose"], help="Enable verbose mode")
    include_dirs = cli.SwitchAttr("-I", list=True, help="Specify include directories")

    @cli.switch("-loglevel", int)
    def set_log_level(self, level):
        """Sets the log-level of the logger"""
        logging.root.setLevel(level)

    def main(self, *srcfiles):
        print("Verbose:", self.verbose)
        print("Include dirs:", self.include_dirs)
        print("Compiling:", srcfiles)


if __name__ == "__main__":
    MyCompiler()



from __future__ import annotations

import pytest

from plumbum import local
from plumbum.cli import Config, ConfigINI

fname = "test_config.ini"


@pytest.mark.usefixtures("cleandir")
class TestConfig:
    def test_makefile(self):
        with ConfigINI(fname) as conf:
            conf["value"] = 12
            conf["string"] = "ho"

        with open(fname) as f:
            contents = f.read()

        assert "value = 12" in contents
        assert "string = ho" in contents

    def test_readfile(self):
        with open(fname, "w") as f:
            print(
                """
[DEFAULT]
one = 1
two = hello""",
                file=f,
            )

        with ConfigINI(fname) as conf:
            assert conf["one"] == "1"
            assert conf["two"] == "hello"

    def test_complex_ini(self):
        with Config(fname) as conf:
            conf["value"] = "normal"
            conf["newer.value"] = "other"

        with Config(fname) as conf:
            assert conf["value"] == "normal"
            assert conf["DEFAULT.value"] == "normal"
            assert conf["newer.value"] == "other"

    def test_nowith(self):
        conf = ConfigINI(fname)
        conf["something"] = "nothing"
        conf.write()

        with open(fname) as f:
            contents = f.read()

        assert "something = nothing" in contents

    def test_home(self):
        mypath = local.env.home / "some_simple_home_rc.ini"
        assert not mypath.exists()
        try:
            with Config("~/some_simple_home_rc.ini") as conf:
                conf["a"] = "b"
            assert mypath.exists()
            mypath.unlink()

            with Config(mypath) as conf:
                conf["a"] = "b"
            assert mypath.exists()
            mypath.unlink()

        finally:
            mypath.unlink()

    def test_notouch(self):
        ConfigINI(fname)
        assert not local.path(fname).exists()

    def test_only_string(self):
        conf = ConfigINI(fname)
        value = conf.get("value", 2)
        assert value == "2"



from __future__ import annotations

import builtins
import io
import itertools
import operator
import os
import typing
import warnings
from abc import ABC, abstractmethod
from functools import reduce

FLAGS = {"f": os.F_OK, "w": os.W_OK, "r": os.R_OK, "x": os.X_OK}


class FSUser(int):
    """A special object that represents a file-system user. It derives from ``int``, so it behaves
    just like a number (``uid``/``gid``), but also have a ``.name`` attribute that holds the
    string-name of the user, if given (otherwise ``None``)
    """

    def __new__(cls, val, name=None):
        self = int.__new__(cls, val)
        self.name = name
        return self


_PathImpl = typing.TypeVar("_PathImpl", bound="Path")


class Path(str, ABC):
    """An abstraction over file system paths. This class is abstract, and the two implementations
    are :class:`LocalPath <plumbum.machines.local.LocalPath>` and
    :class:`RemotePath <plumbum.path.remote.RemotePath>`.
    """

    CASE_SENSITIVE = True

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def __truediv__(self: _PathImpl, other: typing.Any) -> _PathImpl:
        """Joins two paths"""
        return self.join(other)

    def __getitem__(self, key):
        if type(key) == str or isinstance(key, Path):
            return self / key
        return str(self)[key]

    def __floordiv__(self, expr):
        """Returns a (possibly empty) list of paths that matched the glob-pattern under this path"""
        return self.glob(expr)

    def __iter__(self):
        """Iterate over the files in this directory"""
        return iter(self.list())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Path):
            return self._get_info() == other._get_info()
        if isinstance(other, str):
            if self.CASE_SENSITIVE:
                return str(self) == other

            return str(self).lower() == other.lower()

        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        return str(self) > str(other)

    def __ge__(self, other):
        return str(self) >= str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __hash__(self):
        return hash(str(self)) if self.CASE_SENSITIVE else hash(str(self).lower())

    def __bool__(self):
        return bool(str(self))

    def __fspath__(self):
        """Added for Python 3.6 support"""
        return str(self)

    def __contains__(self, item):
        """Paths should support checking to see if an file or folder is in them."""
        try:
            return (self / item.name).exists()
        except AttributeError:
            return (self / item).exists()

    @abstractmethod
    def _form(self: _PathImpl, *parts: typing.Any) -> _PathImpl:
        pass

    def up(self, count=1):
        """Go up in ``count`` directories (the default is 1)"""
        return self.join("../" * count)

    def walk(
        self,
        filter=lambda _: True,  # pylint: disable=redefined-builtin
        dir_filter=lambda _: True,
    ):
        """traverse all (recursive) sub-elements under this directory, that match the given filter.
        By default, the filter accepts everything; you can provide a custom filter function that
        takes a path as an argument and returns a boolean

        :param filter: the filter (predicate function) for matching results. Only paths matching
                       this predicate are returned. Defaults to everything.
        :param dir_filter: the filter (predicate function) for matching directories. Only directories
                           matching this predicate are recursed into. Defaults to everything.
        """
        for p in self.list():
            if filter(p):
                yield p
            if p.is_dir() and dir_filter(p):
                yield from p.walk(filter, dir_filter)

    @property
    @abstractmethod
    def name(self) -> str:
        """The basename component of this path"""

    @property
    def basename(self):
        """Included for compatibility with older Plumbum code"""
        warnings.warn("Use .name instead", FutureWarning, stacklevel=2)
        return self.name

    @property
    @abstractmethod
    def stem(self) -> str:
        """The name without an extension, or the last component of the path"""

    @property
    @abstractmethod
    def dirname(self: _PathImpl) -> _PathImpl:
        """The dirname component of this path"""

    @property
    @abstractmethod
    def root(self) -> str:
        """The root of the file tree (`/` on Unix)"""

    @property
    @abstractmethod
    def drive(self) -> str:
        """The drive letter (on Windows)"""

    @property
    @abstractmethod
    def suffix(self) -> str:
        """The suffix of this file"""

    @property
    @abstractmethod
    def suffixes(self) -> list[str]:
        """This is a list of all suffixes"""

    @property
    @abstractmethod
    def uid(self) -> FSUser:
        """The user that owns this path. The returned value is a :class:`FSUser <plumbum.path.FSUser>`
        object which behaves like an ``int`` (as expected from ``uid``), but it also has a ``.name``
        attribute that holds the string-name of the user"""

    @property
    @abstractmethod
    def gid(self) -> FSUser:
        """The group that owns this path. The returned value is a :class:`FSUser <plumbum.path.FSUser>`
        object which behaves like an ``int`` (as expected from ``gid``), but it also has a ``.name``
        attribute that holds the string-name of the group"""

    @abstractmethod
    def as_uri(self, scheme: str | None = None) -> str:
        """Returns a universal resource identifier. Use ``scheme`` to force a scheme."""

    @abstractmethod
    def _get_info(self) -> typing.Any:
        pass

    @abstractmethod
    def join(self: _PathImpl, *parts: typing.Any) -> _PathImpl:
        """Joins this path with any number of paths"""

    @abstractmethod
    def list(self: _PathImpl) -> builtins.list[_PathImpl]:
        """Returns the files in this directory"""

    @abstractmethod
    def iterdir(self: _PathImpl) -> typing.Iterable[_PathImpl]:
        """Returns an iterator over the directory. Might be slightly faster on Python 3.5 than .list()"""

    @abstractmethod
    def is_dir(self) -> bool:
        """Returns ``True`` if this path is a directory, ``False`` otherwise"""

    def isdir(self):
        """Included for compatibility with older Plumbum code"""
        warnings.warn("Use .is_dir() instead", FutureWarning, stacklevel=2)
        return self.is_dir()

    @abstractmethod
    def is_file(self) -> bool:
        """Returns ``True`` if this path is a regular file, ``False`` otherwise"""

    def isfile(self) -> bool:
        """Included for compatibility with older Plumbum code"""
        warnings.warn("Use .is_file() instead", FutureWarning, stacklevel=2)
        return self.is_file()

    def islink(self):
        """Included for compatibility with older Plumbum code"""
        warnings.warn("Use is_symlink instead", FutureWarning, stacklevel=2)
        return self.is_symlink()

    @abstractmethod
    def is_symlink(self) -> bool:
        """Returns ``True`` if this path is a symbolic link, ``False`` otherwise"""

    @abstractmethod
    def exists(self) -> bool:
        """Returns ``True`` if this path exists, ``False`` otherwise"""

    @abstractmethod
    def stat(self) -> os.stat_result:
        """Returns the os.stats for a file"""

    @abstractmethod
    def with_name(self: _PathImpl, name: typing.Any) -> _PathImpl:
        """Returns a path with the name replaced"""

    @abstractmethod
    def with_suffix(self: _PathImpl, suffix: str, depth: int | None = 1) -> _PathImpl:
        """Returns a path with the suffix replaced. Up to last ``depth`` suffixes will be
        replaced. None will replace all suffixes. If there are less than ``depth`` suffixes,
        this will replace all suffixes. ``.tar.gz`` is an example where ``depth=2`` or
        ``depth=None`` is useful"""

    def preferred_suffix(self, suffix):
        """Adds a suffix if one does not currently exist (otherwise, no change). Useful
        for loading files with a default suffix"""
        return self if len(self.suffixes) > 0 else self.with_suffix(suffix)

    @abstractmethod
    def glob(
        self: _PathImpl, pattern: str | typing.Iterable[str]
    ) -> builtins.list[_PathImpl]:
        """Returns a (possibly empty) list of paths that matched the glob-pattern under this path"""

    @abstractmethod
    def delete(self):
        """Deletes this path (recursively, if a directory)"""

    @abstractmethod
    def move(self, dst):
        """Moves this path to a different location"""

    def rename(self, newname):
        """Renames this path to the ``new name`` (only the basename is changed)"""
        return self.move(self.up() / newname)

    @abstractmethod
    def copy(self, dst, override=None):
        """Copies this path (recursively, if a directory) to the destination path "dst".
        Raises TypeError if dst exists and override is False.
        Will overwrite if override is True.
        Will silently fail to copy if override is None (the default)."""

    @abstractmethod
    def mkdir(self, mode=0o777, parents=True, exist_ok=True):
        """
        Creates a directory at this path.

        :param mode: **Currently only implemented for local paths!** Numeric mode to use for directory
                     creation, which may be ignored on some systems. The current implementation
                     reproduces the behavior of ``os.mkdir`` (i.e., the current umask is first masked
                     out), but this may change for remote paths. As with ``os.mkdir``, it is recommended
                     to call :func:`chmod` explicitly if you need to be sure.
        :param parents: If this is true (the default), the directory's parents will also be created if
                        necessary.
        :param exist_ok: If this is true (the default), no exception will be raised if the directory
                         already exists (otherwise ``OSError``).

        Note that the defaults for ``parents`` and ``exist_ok`` are the opposite of what they are in
        Python's own ``pathlib`` - this is to maintain backwards-compatibility with Plumbum's behaviour
        from before they were implemented.
        """

    @abstractmethod
    def open(self, mode: str = "r", *, encoding: str | None = None) -> io.IOBase:
        """opens this path as a file"""

    @abstractmethod
    def read(self, encoding: str | None = None) -> str:
        """returns the contents of this file as a ``str``. By default the data is read
        as text, but you can specify the encoding, e.g., ``'latin1'`` or ``'utf8'``"""

    @abstractmethod
    def write(self, data: str | bytes, encoding: str | None = None) -> None:
        """writes the given data to this file. By default the data is written as-is
        (either text or binary), but you can specify the encoding, e.g., ``'latin1'``
        or ``'utf8'``"""

    @abstractmethod
    def touch(self):
        """Update the access time. Creates an empty file if none exists."""

    @abstractmethod
    def chown(self, owner=None, group=None, recursive=None):
        """Change ownership of this path.

        :param owner: The owner to set (either ``uid`` or ``username``), optional
        :param group: The group to set (either ``gid`` or ``groupname``), optional
        :param recursive: whether to change ownership of all contained files and subdirectories.
                          Only meaningful when ``self`` is a directory. If ``None``, the value
                          will default to ``True`` if ``self`` is a directory, ``False`` otherwise.
        """

    @abstractmethod
    def chmod(self, mode):
        """Change the mode of path to the numeric mode.

        :param mode: file mode as for os.chmod
        """

    @staticmethod
    def _access_mode_to_flags(mode, flags=None):
        if flags is None:
            flags = FLAGS

        if isinstance(mode, str):
            return reduce(operator.or_, [flags[m] for m in mode.lower()], 0)

        return mode

    @abstractmethod
    def access(self, mode: int | str = 0) -> bool:
        """Test file existence or permission bits

        :param mode: a bitwise-or of access bits, or a string-representation thereof:
                     ``'f'``, ``'x'``, ``'r'``, ``'w'`` for ``os.F_OK``, ``os.X_OK``,
                     ``os.R_OK``, ``os.W_OK``
        """

    @abstractmethod
    def link(self, dst):
        """Creates a hard link from ``self`` to ``dst``

        :param dst: the destination path
        """

    @abstractmethod
    def symlink(self, dst):
        """Creates a symbolic link from ``self`` to ``dst``

        :param dst: the destination path
        """

    @abstractmethod
    def unlink(self):
        """Deletes a symbolic link"""

    def split(self, *_args, **_kargs):
        """Splits the path on directory separators, yielding a list of directories, e.g,
        ``"/var/log/messages"`` will yield ``['var', 'log', 'messages']``.
        """
        parts = []
        path = self
        while path != path.dirname:
            parts.append(path.name)
            path = path.dirname
        return parts[::-1]

    @property
    def parts(self):
        """Splits the directory into parts, including the base directory, returns a tuple"""
        return (self.drive + self.root, *self.split())

    def relative_to(self, source):
        """Computes the "relative path" require to get from ``source`` to ``self``. They satisfy the invariant
        ``source_path + (target_path - source_path) == target_path``. For example::

            /var/log/messages - /var/log/messages = []
            /var/log/messages - /var              = [log, messages]
            /var/log/messages - /                 = [var, log, messages]
            /var/log/messages - /var/tmp          = [.., log, messages]
            /var/log/messages - /opt              = [.., var, log, messages]
            /var/log/messages - /opt/lib          = [.., .., var, log, messages]
        """
        if isinstance(source, str):
            source = self._form(source)
        parts = self.split()
        baseparts = source.split()
        ancestors = len(
            list(itertools.takewhile(lambda p: p[0] == p[1], zip(parts, baseparts)))
        )
        return RelativePath([".."] * (len(baseparts) - ancestors) + parts[ancestors:])

    def __sub__(self, other):
        """Same as ``self.relative_to(other)``"""
        return self.relative_to(other)

    @staticmethod
    def _glob(pattern, fn):
        """Applies a glob string or list/tuple/iterable to the current path, using ``fn``"""
        if isinstance(pattern, str):
            return fn(pattern)

        results = {value for single_pattern in pattern for value in fn(single_pattern)}
        return sorted(results)

    def resolve(self, strict=False):  # noqa: ARG002
        """Added to allow pathlib like syntax. Does nothing since
        Plumbum paths are always absolute. Does not (currently) resolve
        symlinks."""
        # TODO: Resolve symlinks here
        return self

    @property
    def parents(self):
        """Pathlib like sequence of ancestors"""
        as_list = (
            reduce(lambda x, y: self._form(x) / y, self.parts[:i], self.parts[0])
            for i in range(len(self.parts) - 1, 0, -1)
        )
        return tuple(as_list)

    @property
    def parent(self):
        """Pathlib like parent of the path."""
        return self.parents[0]


class RelativePath:
    """
    Relative paths are the "delta" required to get from one path to another.
    Note that relative path do not point at anything, and thus are not paths.
    Therefore they are system agnostic (but closed under addition)
    Paths are always absolute and point at "something", whether existent or not.

    Relative paths are created by subtracting paths (``Path.relative_to``)
    """

    def __init__(self, parts):
        self.parts = parts

    def __str__(self):
        return "/".join(self.parts)

    def __iter__(self):
        return iter(self.parts)

    def __len__(self):
        return len(self.parts)

    def __getitem__(self, index):
        return self.parts[index]

    def __repr__(self):
        return f"RelativePath({self.parts!r})"

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        return str(self) > str(other)

    def __ge__(self, other):
        return str(self) >= str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __hash__(self):
        return hash(str(self))

    def __bool__(self):
        return bool(str(self))

    def up(self, count=1):
        return RelativePath(self.parts[:-count])

    def __radd__(self, path):
        return path.join(*self.parts)



from __future__ import annotations

from plumbum.commands.base import (
    ERROUT,
    BaseCommand,
    ConcreteCommand,
    shquote,
    shquote_list,
)
from plumbum.commands.modifiers import (
    BG,
    FG,
    NOHUP,
    RETCODE,
    TEE,
    TF,
    ExecutionModifier,
    Future,
)
from plumbum.commands.processes import (
    CommandNotFound,
    ProcessExecutionError,
    ProcessLineTimedOut,
    ProcessTimedOut,
    run_proc,
)

__all__ = (
    "BG",
    "ERROUT",
    "FG",
    "NOHUP",
    "RETCODE",
    "TEE",
    "TF",
    "BaseCommand",
    "CommandNotFound",
    "ConcreteCommand",
    "ExecutionModifier",
    "Future",
    "ProcessExecutionError",
    "ProcessLineTimedOut",
    "ProcessTimedOut",
    "run_proc",
    "shquote",
    "shquote_list",
)


def __dir__():
    return __all__



from __future__ import annotations

import pytest

from plumbum import SshMachine, local
from plumbum._testtools import skip_on_windows
from plumbum.path.utils import copy, delete, move


@skip_on_windows
@pytest.mark.ssh
def test_copy_move_delete():
    from plumbum.cmd import touch

    with local.tempdir() as dir:
        (dir / "orog").mkdir()
        (dir / "orog" / "rec").mkdir()
        for i in range(20):
            touch(dir / "orog" / f"f{i}.txt")
        for i in range(20, 40):
            touch(dir / "orog" / "rec" / f"f{i}.txt")

        move(dir / "orog", dir / "orig")

        s1 = sorted(f.name for f in (dir / "orig").walk())

        copy(dir / "orig", dir / "dup")
        s2 = sorted(f.name for f in (dir / "dup").walk())
        assert s1 == s2

        with SshMachine("localhost") as rem, rem.tempdir() as dir2:
            copy(dir / "orig", dir2)
            s3 = sorted(f.name for f in (dir2 / "orig").walk())
            assert s1 == s3

            copy(dir2 / "orig", dir2 / "dup")
            s4 = sorted(f.name for f in (dir2 / "dup").walk())
            assert s1 == s4

            copy(dir2 / "dup", dir / "dup2")
            s5 = sorted(f.name for f in (dir / "dup2").walk())
            assert s1 == s5

            with SshMachine("localhost") as rem2, rem2.tempdir() as dir3:
                copy(dir2 / "dup", dir3)
                s6 = sorted(f.name for f in (dir3 / "dup").walk())
                assert s1 == s6

                move(dir3 / "dup", dir / "superdup")
                assert not (dir3 / "dup").exists()

                s7 = sorted(f.name for f in (dir / "superdup").walk())
                assert s1 == s7

                # test rm
                delete(dir)



"""
This file provides two classes, `Color` and `Style`.

``Color`` is rarely used directly,
but merely provides the workhorse for finding and manipulating colors.

With the ``Style`` class, any color can be directly called or given to a with statement.
"""

from __future__ import annotations

import contextlib
import os
import platform
import re
import sys
from abc import ABCMeta, abstractmethod
from copy import copy
from typing import IO, ClassVar

from .names import (
    FindNearest,
    attributes_ansi,
    color_codes_simple,
    color_html,
    color_names,
    from_html,
)

__all__ = [
    "ANSIStyle",
    "AttributeNotFound",
    "Color",
    "ColorNotFound",
    "HTMLStyle",
    "Style",
]

_lower_camel_names = [n.replace("_", "") for n in color_names]


def get_color_repr():
    """Gets best colors for current system."""
    if "NO_COLOR" in os.environ:
        return 0
    if os.environ.get("FORCE_COLOR", "") in {"0", "1", "2", "3", "4"}:
        return int(os.environ["FORCE_COLOR"])
    if not sys.stdout.isatty():
        return 0

    term = os.environ.get("TERM", "")

    # Some terminals set TERM=xterm for compatibility
    if term.endswith("256color") or term == "xterm":
        return 3 if platform.system() == "Darwin" else 4
    if term.endswith("16color"):
        return 2
    if term == "screen":
        return 1
    if os.name == "nt":
        return 0

    return 3


class ColorNotFound(Exception):
    """Thrown when a color is not valid for a particular method."""


class AttributeNotFound(Exception):
    """Similar to color not found, only for attributes."""


class ResetNotSupported(Exception):
    """An exception indicating that Reset is not available
    for this Style."""


class Color:
    """\
    Loaded with ``(r, g, b, fg)`` or ``(color, fg=fg)``. The second signature is a short cut
    and will try full and hex loading.

    This class stores the idea of a color, rather than a specific implementation.
    It provides as many different tools for representations as possible, and can be subclassed
    to add more representations, though that should not be needed for most situations. ``.from_`` class methods provide quick ways to create colors given different representations.
    You will not usually interact with this class.

    Possible colors::

        reset = Color() # The reset color by default
        background_reset = Color(fg=False) # Can be a background color
        blue = Color(0,0,255) # Red, Green, Blue
        green = Color.from_full("green") # Case insensitive name, from large colorset
        red = Color.from_full(1) # Color number
        white = Color.from_html("#FFFFFF") # HTML supported
        yellow = Color.from_simple("red") # Simple colorset


    The attributes are:

    .. data:: reset

        True it this is a reset color (following attributes don't matter if True)

    .. data:: rgb

        The red/green/blue tuple for this color

    .. data:: simple

        If true will stay to 16 color mode.

    .. data:: number

        The color number given the mode, closest to rgb
        if not rgb not exact, gives position of closest name.

    .. data:: fg

        This is a foreground color if True. Background color if False.

        """

    __slots__ = ("exact", "fg", "isreset", "number", "representation", "rgb")

    def __init__(self, r_or_color=None, g=None, b=None, fg=True):
        """This works from color values, or tries to load non-simple ones."""

        if isinstance(r_or_color, type(self)):
            for item in ("fg", "isreset", "rgb", "number", "representation", "exact"):
                setattr(self, item, getattr(r_or_color, item))
            return

        self.fg = fg
        self.isreset = True  # Starts as reset color
        self.rgb = (0, 0, 0)

        self.number = None
        "Number of the original color, or closest color"

        self.representation = 4
        "0 for off, 1 for 8 colors, 2 for 16 colors, 3 for 256 colors, 4 for true color"

        self.exact = True
        "This is false if the named color does not match the real color"

        if None in (g, b):
            if not r_or_color:
                return
            try:
                self._from_simple(r_or_color)
            except ColorNotFound:
                try:
                    self._from_full(r_or_color)
                except ColorNotFound:
                    self._from_hex(r_or_color)

        elif None not in (r_or_color, g, b):
            self.rgb = (r_or_color, g, b)
            self._init_number()
        else:
            raise ColorNotFound("Invalid parameters for a color!")

    def _init_number(self):
        """Should always be called after filling in r, g, b, and representation.
        Color will not be a reset color anymore."""

        if self.representation in (0, 1):
            number = FindNearest(*self.rgb).only_basic()
        elif self.representation == 2:
            number = FindNearest(*self.rgb).only_simple()
        elif self.representation in (3, 4):
            number = FindNearest(*self.rgb).all_fast()
        else:
            raise AssertionError("Invalid representation, needs to be 0-4")

        if self.number is None:
            self.number = number

        self.isreset = False
        self.exact = self.rgb == from_html(color_html[self.number])
        if not self.exact:
            self.number = number

    @classmethod
    def from_simple(cls, color, fg=True):
        """Creates a color from simple name or color number"""
        self = cls(fg=fg)
        self._from_simple(color)
        return self

    def _from_simple(self, color):
        with contextlib.suppress(AttributeError):
            color = color.lower()
            color = color.replace(" ", "")
            color = color.replace("_", "")

        if color == "reset":
            return

        if color in _lower_camel_names[:16]:
            self.number = _lower_camel_names.index(color)
            self.rgb = from_html(color_html[self.number])

        elif isinstance(color, int) and 0 <= color < 16:
            self.number = color
            self.rgb = from_html(color_html[color])

        else:
            raise ColorNotFound("Did not find color: " + repr(color))

        self.representation = 2
        self._init_number()

    @classmethod
    def from_full(cls, color, fg=True):
        """Creates a color from full name or color number"""
        self = cls(fg=fg)
        self._from_full(color)
        return self

    def _from_full(self, color):
        with contextlib.suppress(AttributeError):
            color = color.lower()
            color = color.replace(" ", "")
            color = color.replace("_", "")

        if color == "reset":
            return

        if color in _lower_camel_names:
            self.number = _lower_camel_names.index(color)
            self.rgb = from_html(color_html[self.number])

        elif isinstance(color, int) and 0 <= color <= 255:
            self.number = color
            self.rgb = from_html(color_html[color])

        else:
            raise ColorNotFound("Did not find color: " + repr(color))

        self.representation = 3
        self._init_number()

    @classmethod
    def from_hex(cls, color, fg=True):
        """Converts #123456 values to colors."""

        self = cls(fg=fg)
        self._from_hex(color)
        return self

    def _from_hex(self, color):
        try:
            self.rgb = from_html(color)
        except (TypeError, ValueError):
            raise ColorNotFound("Did not find htmlcode: " + repr(color)) from None

        self.representation = 4
        self._init_number()

    @property
    def name(self):
        """The (closest) name of the current color"""
        return "reset" if self.isreset else color_names[self.number]

    @property
    def name_camelcase(self):
        """The camelcase name of the color"""
        return self.name.replace("_", " ").title().replace(" ", "")

    def __repr__(self):
        """This class has a smart representation that shows name and color (if not unique)."""
        name = ["Deactivated:", " Basic:", "", " Full:", " True:"][self.representation]
        name += "" if self.fg else " Background"
        name += " " + self.name_camelcase
        name += "" if self.exact else " " + self.hex_code
        return name[1:]

    def __eq__(self, other):
        """Reset colors are equal, otherwise rgb have to match."""
        return other.isreset if self.isreset else self.rgb == other.rgb

    def __hash__(self):
        return hash(self.isreset or self.rgb)

    @property
    def ansi_sequence(self):
        """This is the ansi sequence as a string, ready to use."""
        return "\033[" + ";".join(map(str, self.ansi_codes)) + "m"

    @property
    def ansi_codes(self):
        """This is the full ANSI code, can be reset, simple, 256, or full color."""
        ansi_addition = 30 if self.fg else 40

        if self.isreset:
            return (ansi_addition + 9,)
        if self.representation < 3:
            return (color_codes_simple[self.number] + ansi_addition,)
        if self.representation == 3:
            return (ansi_addition + 8, 5, self.number)

        return (ansi_addition + 8, 2, self.rgb[0], self.rgb[1], self.rgb[2])

    @property
    def hex_code(self):
        """This is the hex code of the current color, html style notation."""

        return (
            "#000000"
            if self.isreset
            else f"#{self.rgb[0]:02X}{self.rgb[1]:02X}{self.rgb[2]:02X}"
        )

    def __str__(self):
        """This just prints it's simple name"""
        return self.name

    def to_representation(self, val):
        """Converts a color to any representation"""
        other = copy(self)
        other.representation = val
        if self.isreset:
            return other
        other.number = None
        other._init_number()
        return other

    def limit_representation(self, val):
        """Only converts if val is lower than representation"""

        return self if self.representation <= val else self.to_representation(val)


class Style(metaclass=ABCMeta):
    """This class allows the color changes to be called directly
    to write them to stdout, ``[]`` calls to wrap colors (or the ``.wrap`` method)
    and can be called in a with statement.
    """

    __slots__ = ("__weakref__", "attributes", "bg", "fg", "isreset")

    color_class = Color
    """The class of color to use. Never hardcode ``Color`` call when writing a Style
    method."""

    # These must be defined by subclasses
    # pylint: disable-next=declare-non-slot
    attribute_names: ClassVar[dict[str, str] | dict[str, int]]

    _stdout: IO | None = None
    end = "\n"
    """The endline character. Override if needed in subclasses."""

    ANSI_REG = re.compile("\033\\[([\\d;]+)m")
    """The regular expression that finds ansi codes in a string."""

    @property
    def stdout(self):
        """\
        This property will allow custom, class level control of stdout.
        It will use current sys.stdout if set to None (default).
        Unfortunately, it only works on an instance..
        """
        # Import sys repeated here to make calling this stable in atexit function
        import sys  # pylint: disable=reimported, redefined-outer-name, import-outside-toplevel

        return (
            self.__class__._stdout if self.__class__._stdout is not None else sys.stdout
        )

    @stdout.setter
    def stdout(self, newout):
        self.__class__._stdout = newout

    def __init__(self, attributes=None, fgcolor=None, bgcolor=None, reset=False):
        """This is usually initialized from a factory."""
        if isinstance(attributes, type(self)):
            for item in ("attributes", "fg", "bg", "isreset"):
                setattr(self, item, copy(getattr(attributes, item)))
            return
        self.attributes = attributes if attributes is not None else {}
        self.fg = fgcolor
        self.bg = bgcolor
        self.isreset = reset
        invalid_attributes = set(self.attributes) - set(self.attribute_names)
        if len(invalid_attributes) > 0:
            raise AttributeNotFound(
                "Attribute(s) not valid: " + ", ".join(invalid_attributes)
            )

    @classmethod
    def from_color(cls, color):
        return cls(fgcolor=color) if color.fg else cls(bgcolor=color)

    def invert(self):
        """This resets current color(s) and flips the value of all
        attributes present"""

        other = self.__class__()

        # Opposite of reset is reset
        if self.isreset:
            other.isreset = True
            return other

        # Flip all attributes
        for attribute in self.attributes:
            other.attributes[attribute] = not self.attributes[attribute]

        # Reset only if color present
        if self.fg:
            other.fg = self.fg.__class__()

        if self.bg:
            other.bg = self.bg.__class__()

        return other

    @property
    def reset(self):
        """Shortcut to access reset as a property."""
        return self.invert()

    def __copy__(self):
        """Copy is supported, will make dictionary and colors unique."""
        result = self.__class__()
        result.isreset = self.isreset
        result.fg = copy(self.fg)
        result.bg = copy(self.bg)
        result.attributes = copy(self.attributes)
        return result

    def __invert__(self):
        """This allows ~color."""
        return self.invert()

    def __add__(self, other):
        """Adding two matching Styles results in a new style with
        the combination of both. Adding with a string results in
        the string concatenation of a style.

        Addition is non-commutative, with the rightmost Style property
        being taken if both have the same property.
        (Not safe)"""
        if type(self) == type(other):
            result = copy(other)

            result.isreset = self.isreset or other.isreset
            for attribute in self.attributes:
                if attribute not in result.attributes:
                    result.attributes[attribute] = self.attributes[attribute]
            if not result.fg:
                result.fg = self.fg
            if not result.bg:
                result.bg = self.bg
            return result

        return other.__class__(self) + other

    def __radd__(self, other):
        """This only gets called if the string is on the left side. (Not safe)"""
        return other + other.__class__(self)

    def wrap(self, wrap_this):
        """Wrap a string in this style and its inverse."""
        return self + wrap_this + ~self

    def __and__(self, other):
        """This class supports ``color & color2`` syntax,
        and ``color & "String" syntax too.``"""
        if type(self) == type(other):
            return self + other

        return self.wrap(other)

    def __rand__(self, other):
        """This class supports ``"String:" & color`` syntax."""
        return self.wrap(other)

    def __ror__(self, other):
        """Support for "String" | color syntax"""
        return self.wrap(other)

    def __or__(self, other):
        """This class supports ``color | color2`` syntax. It also supports
        ``"color | "String"`` syntax too."""
        return self.__and__(other)

    def __call__(self):
        """\
        This is a shortcut to print color immediately to the stdout. (Not safe)
        """

        self.now()

    def now(self):
        """Immediately writes color to stdout. (Not safe)"""
        self.stdout.write(str(self))

    def print(self, *printables, **kargs):
        """\
        This acts like print; will print that argument to stdout wrapped
        in Style with the same syntax as the print function in 3.4."""

        end = kargs.get("end", self.end)
        sep = kargs.get("sep", " ")
        file = kargs.get("file", self.stdout)
        flush = kargs.get("flush", False)
        file.write(self.wrap(sep.join(map(str, printables))) + end)
        if flush:
            file.flush()

    print_ = print
    """DEPRECATED: Shortcut from classic Python 2"""

    def __getitem__(self, wrapped):
        """The [] syntax is supported for wrapping"""
        return self.wrap(wrapped)

    def __enter__(self):
        """Context manager support"""
        self.stdout.write(str(self))
        self.stdout.flush()

    def __exit__(self, _type, _value, _traceback):
        """Runs even if exception occurred, does not catch it."""
        self.stdout.write(str(~self))
        self.stdout.flush()
        return False

    @property
    def ansi_codes(self):
        """Generates the full ANSI code sequence for a Style"""

        if self.isreset:
            return [0]

        codes = []
        for attribute in self.attributes:
            if self.attributes[attribute]:
                codes.append(attributes_ansi[attribute])
            else:
                # Fixing bold inverse being 22 instead of 21 on some terminals:
                codes.append(
                    attributes_ansi[attribute] + 20
                    if attributes_ansi[attribute] != 1
                    else 22
                )

        if self.fg:
            codes.extend(self.fg.ansi_codes)

        if self.bg:
            self.bg.fg = False
            codes.extend(self.bg.ansi_codes)

        return codes

    @property
    def ansi_sequence(self):
        """This is the string ANSI sequence."""
        codes = ";".join(str(c) for c in self.ansi_codes)
        return f"\033[{codes}m" if codes else ""

    def __repr__(self):
        name = self.__class__.__name__
        attributes = ", ".join(a for a in self.attributes if self.attributes[a])
        neg_attributes = ", ".join(
            f"-{a}" for a in self.attributes if not self.attributes[a]
        )
        colors = ", ".join(repr(c) for c in (self.fg, self.bg) if c)
        string = (
            "; ".join(s for s in (attributes, neg_attributes, colors) if s) or "empty"
        )
        if self.isreset:
            string = "reset"
        return f"<{name}: {string}>"

    def __eq__(self, other):
        """Equality is true only if reset, or if attributes, fg, and bg match."""
        if type(self) == type(other):
            if self.isreset:
                return other.isreset

            return (
                self.attributes == other.attributes
                and self.fg == other.fg
                and self.bg == other.bg
            )

        return str(self) == other

    __hash__ = None  # type: ignore[assignment]

    @abstractmethod
    def __str__(self):
        """Base Style does not implement a __str__ representation. This is the one
        required method of a subclass."""

    @classmethod
    def from_ansi(cls, ansi_string, filter_resets=False):
        """This generated a style from an ansi string. Will ignore resets if filter_resets is True."""
        result = cls()
        res = cls.ANSI_REG.search(ansi_string)
        for group in res.groups():
            sequence = map(int, group.split(";"))
            result.add_ansi(sequence, filter_resets)
        return result

    def add_ansi(self, sequence, filter_resets=False):
        """Adds a sequence of ansi numbers to the class. Will ignore resets if filter_resets is True."""

        values = iter(sequence)
        try:
            while True:
                value = next(values)
                if value in {38, 48}:
                    fg = value == 38
                    value = next(values)
                    if value == 5:
                        value = next(values)
                        if fg:
                            self.fg = self.color_class.from_full(value)
                        else:
                            self.bg = self.color_class.from_full(value, fg=False)
                    elif value == 2:
                        r = next(values)
                        g = next(values)
                        b = next(values)
                        if fg:
                            self.fg = self.color_class(r, g, b)
                        else:
                            self.bg = self.color_class(r, g, b, fg=False)
                    else:
                        raise ColorNotFound("the value 5 or 2 should follow a 38 or 48")
                elif value == 0:
                    if filter_resets is False:
                        self.isreset = True
                elif value in attributes_ansi.values():
                    for name, att_value in attributes_ansi.items():
                        if value == att_value:
                            self.attributes[name] = True
                elif value in (20 + n for n in attributes_ansi.values()):
                    if filter_resets is False:
                        for name, att_value in attributes_ansi.items():
                            if value == att_value + 20:
                                self.attributes[name] = False
                elif 30 <= value <= 37:
                    self.fg = self.color_class.from_simple(value - 30)
                elif 40 <= value <= 47:
                    self.bg = self.color_class.from_simple(value - 40, fg=False)
                elif 90 <= value <= 97:
                    self.fg = self.color_class.from_simple(value - 90 + 8)
                elif 100 <= value <= 107:
                    self.bg = self.color_class.from_simple(value - 100 + 8, fg=False)
                elif value == 39:
                    if filter_resets is False:
                        self.fg = self.color_class()
                elif value == 49:
                    if filter_resets is False:
                        self.bg = self.color_class(fg=False)
                else:
                    raise ColorNotFound(f"The code {value} is not recognised")
        except StopIteration:
            pass

    @classmethod
    def string_filter_ansi(cls, colored_string):
        """Filters out colors in a string, returning only the name."""
        return cls.ANSI_REG.sub("", colored_string)

    @classmethod
    def string_contains_colors(cls, colored_string):
        """Checks to see if a string contains colors."""
        return len(cls.ANSI_REG.findall(colored_string)) > 0

    def to_representation(self, rep):
        """This converts both colors to a specific representation"""
        other = copy(self)
        if other.fg:
            other.fg = other.fg.to_representation(rep)
        if other.bg:
            other.bg = other.bg.to_representation(rep)
        return other

    def limit_representation(self, rep):
        """This only converts if true representation is higher"""

        if rep is True or rep is False:
            return self

        other = copy(self)
        if other.fg:
            other.fg = other.fg.limit_representation(rep)
        if other.bg:
            other.bg = other.bg.limit_representation(rep)
        return other

    @property
    def basic(self):
        """The color in the 8 color representation."""
        return self.to_representation(1)

    @property
    def simple(self):
        """The color in the 16 color representation."""
        return self.to_representation(2)

    @property
    def full(self):
        """The color in the 256 color representation."""
        return self.to_representation(3)

    @property
    def true(self):
        """The color in the true color representation."""
        return self.to_representation(4)


class ANSIStyle(Style):
    """This is a subclass for ANSI styles. Use it to get
    color on sys.stdout tty terminals on posix systems.

    Set ``use_color = True/False`` if you want to control color
    for anything using this Style."""

    __slots__ = ()
    use_color = get_color_repr()

    attribute_names = attributes_ansi

    def __str__(self):
        return (
            self.limit_representation(self.use_color).ansi_sequence
            if self.use_color
            else ""
        )


class HTMLStyle(Style):
    """This was meant to be a demo of subclassing Style, but
    actually can be a handy way to quickly color html text."""

    __slots__ = ()
    attribute_names = {
        "bold": "b",
        "em": "em",
        "italics": "i",
        "li": "li",
        "underline": 'span style="text-decoration: underline;"',
        "code": "code",
        "ol": "ol start=0",
        "strikeout": "s",
    }
    end = "<br/>\n"

    def __str__(self):
        if self.isreset:
            raise ResetNotSupported("HTML does not support global resets!")

        result = ""

        if self.bg and not self.bg.isreset:
            result += f'<span style="background-color: {self.bg.hex_code}">'
        if self.fg and not self.fg.isreset:
            result += f'<font color="{self.fg.hex_code}">'
        for attr in sorted(self.attributes):
            if self.attributes[attr]:
                result += "<" + self.attribute_names[attr] + ">"

        for attr in sorted(self.attributes, reverse=True):
            if not self.attributes[attr]:
                result += "</" + self.attribute_names[attr].split(" ")[0] + ">"
        if self.fg and self.fg.isreset:
            result += "</font>"
        if self.bg and self.bg.isreset:
            result += "</span>"

        return result



from __future__ import annotations

from plumbum import cli


class TestValidator:
    def test_named(self):
        class Try:
            @cli.positional(x=abs, y=str)
            def main(selfy, x, y):
                pass

        assert Try.main.positional == [abs, str]
        assert Try.main.positional_varargs is None

    def test_position(self):
        class Try:
            @cli.positional(abs, str)
            def main(selfy, x, y):
                pass

        assert Try.main.positional == [abs, str]
        assert Try.main.positional_varargs is None

    def test_mix(self):
        class Try:
            @cli.positional(abs, str, d=bool)
            def main(selfy, x, y, z, d):
                pass

        assert Try.main.positional == [abs, str, None, bool]
        assert Try.main.positional_varargs is None

    def test_var(self):
        class Try:
            @cli.positional(abs, str, int)
            def main(selfy, x, y, *g):
                pass

        assert Try.main.positional == [abs, str]
        assert Try.main.positional_varargs is int

    def test_defaults(self):
        class Try:
            @cli.positional(abs, str)
            def main(selfy, x, y="hello"):
                pass

        assert Try.main.positional == [abs, str]


class TestProg:
    def test_prog(self, capsys):
        class MainValidator(cli.Application):
            @cli.positional(int, int, int)
            def main(self, myint, myint2, *mylist):
                print(repr(myint), myint2, mylist)

        _, rc = MainValidator.run(["prog", "1", "2", "3", "4", "5"], exit=False)
        assert rc == 0
        assert capsys.readouterr()[0].strip() == "1 2 (3, 4, 5)"

    def test_failure(self, capsys):
        class MainValidator(cli.Application):
            @cli.positional(int, int, int)
            def main(self, myint, myint2, *mylist):
                print(myint, myint2, mylist)

        _, rc = MainValidator.run(["prog", "1.2", "2", "3", "4", "5"], exit=False)

        assert rc == 2
        value = capsys.readouterr()[0].strip()
        assert "int" in value
        assert "not" in value
        assert "1.2" in value

    def test_defaults(self, capsys):
        class MainValidator(cli.Application):
            @cli.positional(int, int)
            def main(self, myint, myint2=2):
                print(repr(myint), repr(myint2))

        _, rc = MainValidator.run(["prog", "1"], exit=False)
        assert rc == 0
        assert capsys.readouterr()[0].strip() == "1 2"

        _, rc = MainValidator.run(["prog", "1", "3"], exit=False)
        assert rc == 0
        assert capsys.readouterr()[0].strip() == "1 3"



from __future__ import annotations

import sys
from collections import OrderedDict
from contextlib import contextmanager
from io import StringIO

from plumbum.cli.terminal import Progress, ask, choose, hexdump, prompt


@contextmanager
def send_stdin(stdin="\n"):
    prevstdin = sys.stdin
    sys.stdin = StringIO(stdin)
    try:
        yield sys.stdin
    finally:
        sys.stdin = prevstdin


class TestPrompt:
    def test_simple(self, capsys):
        with send_stdin("12"):
            assert prompt("Enter a random int:", type=int) == 12
        assert capsys.readouterr()[0] == "Enter a random int: "

    def test_try_twice(self, capsys):
        with send_stdin("\n13"):
            assert prompt("Enter a random int:", type=int) == 13
        assert capsys.readouterr()[0] == "Enter a random int: Enter a random int: "

    def test_str(self):
        with send_stdin("1234"):
            assert prompt("Enter a string", type=str) == "1234"

    def test_default(self, capsys):
        with send_stdin(""):
            assert prompt("Enter nothing", default="hi") == "hi"
        assert capsys.readouterr()[0] == "Enter nothing [hi]: "

    def test_typefail(self, capsys):
        with send_stdin("1.2\n13"):
            assert prompt("Enter int", type=int) == 13
        assert "try again" in capsys.readouterr()[0]

    def test_validator(self, capsys):
        with send_stdin("12\n9"):
            assert (
                prompt("Enter in range < 10", type=int, validator=lambda x: x < 10) == 9
            )
        assert "try again" in capsys.readouterr()[0]


class TestTerminal:
    def test_ask(self, capsys):
        with send_stdin("\n"):
            assert ask("Do you like cats?", default=True)
        assert capsys.readouterr()[0] == "Do you like cats? [Y/n] "

        with send_stdin("\nyes"):
            assert ask("Do you like cats?")
        assert (
            capsys.readouterr()[0]
            == "Do you like cats? (y/n) Invalid response, please try again\nDo you like cats? (y/n) "
        )

    def test_choose(self, capsys):
        with send_stdin("foo\n2\n"):
            assert (
                choose("What is your favorite color?", ["blue", "yellow", "green"])
                == "yellow"
            )
        assert (
            capsys.readouterr()[0]
            == "What is your favorite color?\n(1) blue\n(2) yellow\n(3) green\nChoice: Invalid choice, please try again\nChoice: "
        )

        with send_stdin("foo\n2\n"):
            assert (
                choose(
                    "What is your favorite color?",
                    [("blue", 10), ("yellow", 11), ("green", 12)],
                )
                == 11
            )
        assert (
            capsys.readouterr()[0]
            == "What is your favorite color?\n(1) blue\n(2) yellow\n(3) green\nChoice: Invalid choice, please try again\nChoice: "
        )

        with send_stdin("foo\n\n"):
            assert (
                choose(
                    "What is your favorite color?",
                    ["blue", "yellow", "green"],
                    default="yellow",
                )
                == "yellow"
            )
        assert (
            capsys.readouterr()[0]
            == "What is your favorite color?\n(1) blue\n(2) yellow\n(3) green\nChoice [2]: Invalid choice, please try again\nChoice [2]: "
        )

    def test_choose_dict(self):
        with send_stdin("23\n1"):
            value = choose("Pick", {"one": "a", "two": "b"})
            assert value in ("a", "b")

    def test_ordered_dict(self):
        dic = OrderedDict()
        dic["one"] = "a"
        dic["two"] = "b"
        with send_stdin("1"):
            value = choose("Pick", dic)
            assert value == "a"
        with send_stdin("2"):
            value = choose("Pick", dic)
            assert value == "b"

    def test_choose_dict_default(self, capsys):
        dic = OrderedDict()
        dic["one"] = "a"
        dic["two"] = "b"
        with send_stdin():
            assert choose("Pick", dic, default="a") == "a"
        assert "[1]" in capsys.readouterr()[0]

    def test_hexdump(self):
        data = "hello world my name is queen marry" + "A" * 66 + "foo bar"
        output = """\
000000 | 68 65 6c 6c 6f 20 77 6f 72 6c 64 20 6d 79 20 6e | hello world my n
000010 | 61 6d 65 20 69 73 20 71 75 65 65 6e 20 6d 61 72 | ame is queen mar
000020 | 72 79 41 41 41 41 41 41 41 41 41 41 41 41 41 41 | ryAAAAAAAAAAAAAA
000030 | 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 | AAAAAAAAAAAAAAAA
*
000060 | 41 41 41 41 66 6f 6f 20 62 61 72                | AAAAfoo bar"""
        assert "\n".join(hexdump(data)) == output

        assert "\n".join(hexdump(StringIO(data))) == output

    def test_progress(self, capsys):
        for _ in Progress.range(4, has_output=True, timer=False):
            print("hi")
        stdout, _stderr = capsys.readouterr()
        output = """\
0% complete
0% complete
hi
25% complete
hi
50% complete
hi
75% complete
hi
100% complete

"""
        assert stdout == output

    def test_progress_empty(self, capsys):
        for _ in Progress.range(0, has_output=True, timer=False):
            print("hi")
        stdout = capsys.readouterr().out
        output = "0/0 complete"
        assert output in stdout



#!/usr/bin/env python3

# If you are on macOS and using brew, you might need the following first:
# export PATH="/usr/local/opt/gettext/bin:$PATH"
from __future__ import annotations

from plumbum import FG, local
from plumbum.cmd import msgfmt, msgmerge, xgettext

translation_dir = local.cwd / "plumbum/cli/i18n"
template = translation_dir / "messages.pot"

(
    xgettext[
        "--from-code",
        "utf-8",
        "-L",
        "python",
        "--keyword=T_",
        "--package-name=Plumbum.cli",
        "-o",
        template,
        sorted(x - local.cwd for x in local.cwd / "plumbum/cli" // "*.py"),
    ]
    & FG
)

for translation in translation_dir // "*.po":
    lang = translation.stem
    new_tfile = translation.with_suffix(".po.new")

    # Merge changes to new file
    (msgmerge[translation, template] > new_tfile) & FG

    new_tfile.move(translation)

    # Render new file into runtime output
    local_dir = translation_dir / lang / "LC_MESSAGES"
    if not local_dir.exists():
        local_dir.mkdir()
    msgfmt["-o", local_dir / "plumbum.cli.mo", translation] & FG



from __future__ import annotations

from plumbum.commands.base import BaseCommand
from plumbum.commands.processes import CommandNotFound, ProcessExecutionError, run_proc


def make_concurrent(self, rhs):
    if not isinstance(rhs, BaseCommand):
        raise TypeError("rhs must be an instance of BaseCommand")

    if isinstance(self, ConcurrentCommand):
        if isinstance(rhs, ConcurrentCommand):
            self.commands.extend(rhs.commands)
        else:
            self.commands.append(rhs)
        return self

    if isinstance(rhs, ConcurrentCommand):
        rhs.commands.insert(0, self)
        return rhs

    return ConcurrentCommand(self, rhs)


BaseCommand.__and__ = make_concurrent


class ConcurrentPopen:
    def __init__(self, procs):
        self.procs = procs
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self.custom_encoding = None
        self.returncode = None

    @property
    def argv(self):
        return [getattr(proc, "argv", []) for proc in self.procs]

    def poll(self):
        if self.returncode is not None:
            return self.returncode
        rcs = [proc.poll() for proc in self.procs]
        if any(rc is None for rc in rcs):
            return None
        self.returncode = 0
        for rc in rcs:
            if rc != 0:
                self.returncode = rc
                break
        return self.returncode

    def wait(self):
        for proc in self.procs:
            proc.wait()
        return self.poll()

    def communicate(self, input=None):
        if input:
            raise ValueError("Cannot pass input to ConcurrentPopen.communicate")
        out_err_tuples = [proc.communicate() for proc in self.procs]
        self.wait()
        return tuple(zip(*out_err_tuples))


class ConcurrentCommand(BaseCommand):
    def __init__(self, *commands):
        self.commands = list(commands)

    def formulate(self, level=0, args=()):
        form = ["("]
        for cmd in self.commands:
            form.extend(cmd.formulate(level, args))
            form.append("&")
        return [*form, ")"]

    def popen(self, *args, **kwargs):
        return ConcurrentPopen([cmd[args].popen(**kwargs) for cmd in self.commands])

    def __getitem__(self, args):
        """Creates a bound-command with the given arguments"""
        if not isinstance(args, (tuple, list)):
            args = [
                args,
            ]
        if not args:
            return self

        return ConcurrentCommand(*(cmd[args] for cmd in self.commands))


class Cluster:
    def __init__(self, *machines):
        self.machines = list(machines)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()

    def close(self):
        for mach in self.machines:
            mach.close()
        del self.machines[:]

    def add_machine(self, machine):
        self.machines.append(machine)

    def __iter__(self):
        return iter(self.machines)

    def filter(self, pred):
        return self.__class__(filter(pred, self))

    def which(self, progname):
        return [mach.which(progname) for mach in self]

    def list_processes(self):
        return [mach.list_processes() for mach in self]

    def pgrep(self, pattern):
        return [mach.pgrep(pattern) for mach in self]

    def path(self, *parts):
        return [mach.path(*parts) for mach in self]

    def __getitem__(self, progname):
        if not isinstance(progname, str):
            raise TypeError(
                "progname must be a string, not {!r}".format(
                    type(
                        progname,
                    )
                )
            )
        return ConcurrentCommand(*(mach[progname] for mach in self))

    def __contains__(self, cmd):
        try:
            self[cmd]
        except CommandNotFound:
            return False
        else:
            return True

    @property
    def python(self):
        return ConcurrentCommand(*(mach.python for mach in self))

    def session(self):
        return ClusterSession(*(mach.session() for mach in self))


class ClusterSession:
    def __init__(self, *sessions):
        self.sessions = sessions

    def __iter__(self):
        return iter(self.sessions)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()

    def __del__(self):
        try:  # noqa: SIM105
            self.close()
        except Exception:
            pass

    def alive(self):
        """Returns ``True`` if the underlying shells are all alive, ``False`` otherwise"""
        return all(session.alive for session in self)

    def close(self):
        """Closes (terminates) all underlying shell sessions"""
        for session in self.sessions:
            session.close()
        del self.sessions[:]

    def popen(self, cmd):
        return ConcurrentPopen([session.popen(cmd) for session in self])

    def run(self, cmd, retcode=None):
        return run_proc(self.popen(cmd), retcode)


if __name__ == "__main__":
    from plumbum import local
    from plumbum.cmd import date, ls, sleep

    c = ls & date & sleep[1]
    print(c())

    c = ls & date & sleep[1] & sleep["-z"]
    try:
        c()
    except ProcessExecutionError as ex:
        print(ex)
    else:
        raise AssertionError("Expected an ProcessExecutionError")

    clst = Cluster(local, local, local)
    print(clst["ls"]())

    # This works fine
    print(local.session().run("echo $$"))

    # this does not
    ret, stdout, stderr = clst.session().run("echo $$")
    print(ret)
    ret = [int(pid) for pid in stdout]
    assert len(set(ret)) == 3



from __future__ import annotations

from plumbum.commands.processes import (
    CommandNotFound,
    ProcessExecutionError,
    ProcessTimedOut,
)


class PopenAddons:
    """This adds a verify to popen objects to that the correct command is attributed when
    an error is thrown."""

    def verify(self, retcode, timeout, stdout, stderr):
        """This verifies that the correct command is attributed."""
        if getattr(self, "_timed_out", False):
            raise ProcessTimedOut(
                f"Process did not terminate within {timeout} seconds",
                getattr(self, "argv", None),
            )

        if retcode is not None:
            if hasattr(retcode, "__contains__"):
                if self.returncode not in retcode:
                    raise ProcessExecutionError(
                        getattr(self, "argv", None), self.returncode, stdout, stderr
                    )
            elif self.returncode != retcode:
                raise ProcessExecutionError(
                    getattr(self, "argv", None), self.returncode, stdout, stderr
                )


class BaseMachine:
    """This is a base class for other machines. It contains common code to
    all machines in Plumbum."""

    def get(self, cmd, *othercommands):
        """This works a little like the ``.get`` method with dict's, only
        it supports an unlimited number of arguments, since later arguments
        are tried as commands and could also fail. It
        will try to call the first command, and if that is not found,
        it will call the next, etc. Will raise if no file named for the
        executable if a path is given, unlike ``[]`` access.

        Usage::

            best_zip = local.get('pigz','gzip')
        """
        try:
            command = self[cmd]
            if not command.executable.exists():
                raise CommandNotFound(cmd, command.executable)
            return command
        except CommandNotFound:
            if othercommands:
                return self.get(othercommands[0], *othercommands[1:])
            raise

    def __contains__(self, cmd):
        """Tests for the existence of the command, e.g., ``"ls" in plumbum.local``.
        ``cmd`` can be anything acceptable by ``__getitem__``.
        """
        try:
            self[cmd]
        except CommandNotFound:
            return False
        return True

    @property
    def encoding(self):
        "This is a wrapper for custom_encoding"
        return self.custom_encoding

    @encoding.setter
    def encoding(self, value):
        self.custom_encoding = value

    def daemonic_popen(self, command, cwd="/", stdout=None, stderr=None, append=True):
        raise NotImplementedError("This is not implemented on this machine!")

    class Cmd:
        def __init__(self, machine):
            self._machine = machine

        def __getattr__(self, name):
            try:
                return self._machine[name]
            except CommandNotFound:
                raise AttributeError(name) from None

    @property
    def cmd(self):
        return self.Cmd(self)

    def clear_program_cache(self):
        """
        Clear the program cache, which is populated via ``machine.which(progname)`` calls.

        This cache speeds up the lookup of a program in the machines PATH, and is particularly
        effective for RemoteMachines.
        """
        self._program_cache.clear()



#!/usr/bin/env python3
from __future__ import annotations

import logging

from plumbum import cli, local
from plumbum.path.utils import copy, delete

logger = logging.getLogger("FileCopier")


class FileCopier(cli.Application):
    overwrite = cli.Flag("-o", help="If given, overwrite existing files")

    @cli.switch(["-l", "--log-to-file"], argtype=str)
    def log_to_file(self, filename):
        """logs all output to the given file"""
        handler = logging.FileHandler(filename)
        logger.addHandler(handler)

    @cli.switch(["--verbose"], requires=["--log-to-file"])
    def set_debug(self):
        """Sets verbose mode"""
        logger.setLevel(logging.DEBUG)

    def main(self, src, dst):
        if local.path(dst).exists():
            if not self.overwrite:
                logger.debug("Oh no! That's terrible")
                raise ValueError("Destination already exists")
            delete(dst)

        logger.debug("I'm going to copy %s to %s", src, dst)
        copy(src, dst)
        logger.debug("Great success")


if __name__ == "__main__":
    FileCopier.run()



from __future__ import annotations

import pytest

import plumbum
from plumbum._testtools import skip_on_windows
from plumbum.commands import BaseCommand


@skip_on_windows
@pytest.mark.timeout(3)
def test_draining_stderr(generate_cmd, process_cmd):
    stdout, stderr = get_output_with_iter_lines(
        generate_cmd | process_cmd | process_cmd
    )
    expected_output = {f"generated {i}" for i in range(5000)}
    expected_output.update(f"consumed {i}" for i in range(5000))
    assert set(stderr) - expected_output == set()
    assert len(stderr) == 15000
    assert len(stdout) == 5000


@skip_on_windows
@pytest.mark.timeout(3)
def test_draining_stderr_with_stderr_redirect(tmp_path, generate_cmd, process_cmd):
    stdout, stderr = get_output_with_iter_lines(
        generate_cmd | (process_cmd >= str(tmp_path / "output.txt")) | process_cmd
    )
    expected_output = {f"generated {i}" for i in range(5000)}
    expected_output.update(f"consumed {i}" for i in range(5000))
    assert set(stderr) - expected_output == set()
    assert len(stderr) == 10000
    assert len(stdout) == 5000


@skip_on_windows
@pytest.mark.timeout(3)
def test_draining_stderr_with_stdout_redirect(tmp_path, generate_cmd, process_cmd):
    stdout, stderr = get_output_with_iter_lines(
        generate_cmd | process_cmd | process_cmd > str(tmp_path / "output.txt")
    )
    expected_output = {f"generated {i}" for i in range(5000)}
    expected_output.update(f"consumed {i}" for i in range(5000))
    assert set(stderr) - expected_output == set()
    assert len(stderr) == 15000
    assert len(stdout) == 0


@pytest.fixture
def generate_cmd(tmp_path):
    generate = tmp_path / "generate.py"
    generate.write_text(
        """\
import sys
for i in range(5000):
    print("generated", i, file=sys.stderr)
    print(i)
"""
    )
    return plumbum.local["python"][generate]


@pytest.fixture
def process_cmd(tmp_path):
    process = tmp_path / "process.py"
    process.write_text(
        """\
import sys
for line in sys.stdin:
    i = line.strip()
    print("consumed", i, file=sys.stderr)
    print(i)
"""
    )
    return plumbum.local["python"][process]


def get_output_with_iter_lines(cmd: BaseCommand) -> tuple[list[str], list[str]]:
    stderr, stdout = [], []
    proc = cmd.popen()
    for stdout_line, stderr_line in proc.iter_lines(retcode=[0, None]):
        if stderr_line:
            stderr.append(stderr_line)
        if stdout_line:
            stdout.append(stdout_line)
    proc.wait()
    return stdout, stderr



"""
This module imitates a real module, providing standard syntax
like from `plumbum.colors` and from `plumbum.colors.bg` to work alongside
all the standard syntax for colors.
"""

from __future__ import annotations

import atexit
import sys

from plumbum.colorlib import ansicolors, main

_reset = ansicolors.reset.now
if __name__ == "__main__":
    main()
else:  # Don't register an exit if this is called using -m!
    atexit.register(_reset)

sys.modules[__name__ + ".fg"] = ansicolors.fg
sys.modules[__name__ + ".bg"] = ansicolors.bg
sys.modules[__name__] = ansicolors  # type: ignore[assignment]



from __future__ import annotations

import contextlib
import errno
import logging
import os
import stat

from plumbum.commands.base import shquote
from plumbum.commands.processes import ProcessLineTimedOut, iter_lines
from plumbum.machines.base import PopenAddons
from plumbum.machines.remote import BaseRemoteMachine
from plumbum.machines.session import ShellSession
from plumbum.path.local import LocalPath
from plumbum.path.remote import RemotePath, StatRes

try:
    # Sigh... we need to gracefully-import paramiko for Sphinx builds, etc
    import paramiko
except ImportError:

    class paramiko:  # type: ignore[no-redef]
        def __bool__(self):
            return False

        def __getattr__(self, name):
            raise ImportError("No module named paramiko")

    paramiko = paramiko()  # type: ignore[operator]

logger = logging.getLogger("plumbum.paramiko")


class ParamikoPopen(PopenAddons):
    def __init__(
        self,
        argv,
        stdin,
        stdout,
        stderr,
        encoding,
        stdin_file=None,
        stdout_file=None,
        stderr_file=None,
    ):
        self.argv = argv
        self.channel = stdout.channel
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.custom_encoding = encoding
        self.returncode = None
        self.pid = None
        self.stdin_file = stdin_file
        self.stdout_file = stdout_file
        self.stderr_file = stderr_file

    def poll(self):
        if self.returncode is None and self.channel.exit_status_ready():
            return self.wait()
        return self.returncode

    def wait(self):
        if self.returncode is None:
            self.channel.recv_exit_status()
            self.returncode = self.channel.exit_status
            self.close()
        return self.returncode

    def close(self):
        self.channel.shutdown_read()
        self.channel.shutdown_write()
        self.channel.close()

    @staticmethod
    def kill():
        # possible way to obtain pid:
        # "(cmd ; echo $?) & echo ?!"
        # and then client.exec_command("kill -9 %s" % (pid,))
        raise OSError("Cannot kill remote processes, we don't have their PIDs")

    terminate = kill

    def send_signal(self, sig):
        raise NotImplementedError()

    def communicate(self):
        stdout = []
        stderr = []
        infile = self.stdin_file
        sources = [
            ("1", stdout, self.stdout, self.stdout_file),
            ("2", stderr, self.stderr, self.stderr_file),
        ]
        i = 0
        while sources:
            if infile:
                try:
                    line = infile.readline()
                except (ValueError, OSError):
                    line = None
                logger.debug("communicate: %r", line)
                if not line:
                    infile.close()
                    infile = None
                    self.stdin.close()
                else:
                    self.stdin.write(line)
                    self.stdin.flush()

            i = (i + 1) % len(sources)
            _name, coll, pipe, outfile = sources[i]
            line = pipe.readline()
            # logger.debug("%s> %r", name, line)
            if not line:
                del sources[i]
            elif outfile:
                outfile.write(line)
                outfile.flush()
            else:
                coll.append(line)
        self.wait()
        stdout = "".join(s for s in stdout).encode(self.custom_encoding)
        stderr = "".join(s for s in stderr).encode(self.custom_encoding)
        return stdout, stderr

    def iter_lines(self, timeout=None, **kwargs):
        if timeout is not None:
            raise NotImplementedError(
                "The 'timeout' parameter is not supported with ParamikoMachine"
            )
        return iter_lines(self, _iter_lines=_iter_lines, **kwargs)

    __iter__ = iter_lines


class ParamikoMachine(BaseRemoteMachine):
    """
    An implementation of :class:`remote machine <plumbum.machines.remote.BaseRemoteMachine>`
    over Paramiko (a Python implementation of openSSH2 client/server). Invoking a remote command
    translates to invoking it over SSH ::

        with ParamikoMachine("yourhostname") as rem:
            r_ls = rem["ls"]
            # r_ls is the remote `ls`
            # executing r_ls() is equivalent to `ssh yourhostname ls`, only without
            # spawning a new ssh client

    :param host: the host name to connect to (SSH server)

    :param user: the user to connect as (if ``None``, the default will be used)

    :param port: the server's port (if ``None``, the default will be used)

    :param password: the user's password (if a password-based authentication is to be performed)
                     (if ``None``, key-based authentication will be used)

    :param keyfile: the path to the identity file (if ``None``, the default will be used)

    :param load_system_host_keys: whether or not to load the system's host keys (from ``/etc/ssh``
                                  and ``~/.ssh``). The default is ``True``, which means Paramiko
                                  behaves much like the ``ssh`` command-line client

    :param missing_host_policy: the value passed to the underlying ``set_missing_host_key_policy``
                                of the client. The default is ``None``, which means
                                ``set_missing_host_key_policy`` is not invoked and paramiko's
                                default behavior (reject) is employed

    :param encoding: the remote machine's encoding (defaults to UTF8)

    :param look_for_keys: set to False to disable searching for discoverable
                          private key files in ``~/.ssh``

    :param connect_timeout: timeout for TCP connection

    .. note:: If Paramiko 1.15 or above is installed, can use GSS_API authentication

    :param bool gss_auth: ``True`` if you want to use GSS-API authentication

    :param bool gss_kex: Perform GSS-API Key Exchange and user authentication

    :param bool gss_deleg_creds: Delegate GSS-API client credentials or not

    :param str gss_host: The targets name in the kerberos database. default: hostname

    :param bool get_pty: Execute remote commands with allocated pseudo-tty. default: False

    :param bool load_system_ssh_config: read system SSH config for ProxyCommand configuration. default: False
    """

    class RemoteCommand(BaseRemoteMachine.RemoteCommand):  # type: ignore[valid-type, misc]
        def __or__(self, *_):
            return NotImplemented

        def __gt__(self, *_):
            return NotImplemented

        def __rshift__(self, *_):
            return NotImplemented

        def __ge__(self, *_):
            return NotImplemented

        def __lt__(self, *_):
            return NotImplemented

        def __lshift__(self, *_):
            return NotImplemented

    def __init__(
        self,
        host,
        user=None,
        port=None,
        password=None,
        keyfile=None,
        load_system_host_keys=True,
        missing_host_policy=None,
        encoding="utf8",
        look_for_keys=None,
        connect_timeout=None,
        keep_alive=0,
        gss_auth=False,
        gss_kex=None,
        gss_deleg_creds=None,
        gss_host=None,
        get_pty=False,
        load_system_ssh_config=False,
    ):
        self.host = host
        kwargs = {}
        if user:
            self._fqhost = f"{user}@{host}"
            kwargs["username"] = user
        else:
            self._fqhost = host
        self._client = paramiko.SSHClient()
        if load_system_host_keys:
            self._client.load_system_host_keys()
        if port is not None:
            kwargs["port"] = port
        if keyfile is not None:
            kwargs["key_filename"] = keyfile
        if password is not None:
            kwargs["password"] = password
        if missing_host_policy is not None:
            self._client.set_missing_host_key_policy(missing_host_policy)
        if look_for_keys is not None:
            kwargs["look_for_keys"] = look_for_keys
        if connect_timeout is not None:
            kwargs["timeout"] = connect_timeout
        if gss_auth:
            kwargs["gss_auth"] = gss_auth
            kwargs["gss_kex"] = gss_kex
            kwargs["gss_deleg_creds"] = gss_deleg_creds
            if not gss_host:
                gss_host = host
            kwargs["gss_host"] = gss_host
        if load_system_ssh_config:
            ssh_config = paramiko.SSHConfig()
            with open(os.path.expanduser("~/.ssh/config"), encoding="utf-8") as f:
                ssh_config.parse(f)
            with contextlib.suppress(KeyError):
                hostConfig = ssh_config.lookup(host)
                kwargs["sock"] = paramiko.ProxyCommand(hostConfig["proxycommand"])
        self._client.connect(host, **kwargs)
        self._keep_alive = keep_alive
        self._sftp = None
        self._get_pty = get_pty
        BaseRemoteMachine.__init__(self, encoding, connect_timeout)

    def __str__(self):
        return f"paramiko://{self._fqhost}"

    def close(self):
        BaseRemoteMachine.close(self)
        self._client.close()

    @property
    def sftp(self):
        """
        Returns an SFTP client on top of the current SSH connection; it can be used to manipulate
        files directly, much like an interactive FTP/SFTP session
        """
        if not self._sftp:
            self._sftp = self._client.open_sftp()
        return self._sftp

    def session(
        self,
        isatty=False,
        term="vt100",
        width=80,
        height=24,
        *,
        new_session=False,  # noqa: ARG002
    ):
        # new_session is ignored for ParamikoMachine
        trans = self._client.get_transport()
        trans.set_keepalive(self._keep_alive)
        chan = trans.open_session()
        if isatty:
            chan.get_pty(term, width, height)
            chan.set_combine_stderr(True)
        chan.invoke_shell()
        stdin = chan.makefile("wb", -1)
        stdout = chan.makefile("rb", -1)
        stderr = chan.makefile_stderr("rb", -1)
        proc = ParamikoPopen(["<shell>"], stdin, stdout, stderr, self.custom_encoding)
        return ShellSession(proc, self.custom_encoding, isatty)

    def popen(
        self,
        args,
        stdin=None,
        stdout=None,
        stderr=None,
        new_session=False,  # noqa: ARG002
        env=None,
        cwd=None,
    ):
        # new_session is ignored for ParamikoMachine
        argv = []
        envdelta = self.env.getdelta()
        if env:
            envdelta.update(env)
        argv.extend(["cd", str(cwd or self.cwd), "&&"])
        if envdelta:
            argv.append("env")
            argv.extend(f"{k}={shquote(v)}" for k, v in envdelta.items())
        argv.extend(args.formulate())
        cmdline = " ".join(argv)
        logger.debug(cmdline)
        si, so, se = self._client.exec_command(cmdline, 1, get_pty=self._get_pty)
        return ParamikoPopen(
            argv,
            si,
            so,
            se,
            self.custom_encoding,
            stdin_file=stdin,
            stdout_file=stdout,
            stderr_file=stderr,
        )

    def download(self, src, dst):
        if isinstance(src, LocalPath):
            raise TypeError(f"src of download cannot be {src!r}")
        if isinstance(src, RemotePath) and src.remote != self:
            raise TypeError(f"src {src!r} points to a different remote machine")
        if isinstance(dst, RemotePath):
            raise TypeError(f"dst of download cannot be {dst!r}")
        return self._download(
            src if isinstance(src, RemotePath) else self.path(src),
            dst if isinstance(dst, LocalPath) else LocalPath(dst),
        )

    def _download(self, src, dst):
        if src.is_dir():
            if not dst.exists():
                self.sftp.mkdir(str(dst))
            for fn in src:
                self._download(fn, dst / fn.name)
        elif dst.is_dir():
            self.sftp.get(str(src), str(dst / src.name))
        else:
            self.sftp.get(str(src), str(dst))

    def upload(self, src, dst):
        if isinstance(src, RemotePath):
            raise TypeError(f"src of upload cannot be {src!r}")
        if isinstance(dst, LocalPath):
            raise TypeError(f"dst of upload cannot be {dst!r}")
        if isinstance(dst, RemotePath) and dst.remote != self:
            raise TypeError(f"dst {dst!r} points to a different remote machine")
        return self._upload(
            src if isinstance(src, LocalPath) else LocalPath(src),
            dst if isinstance(dst, RemotePath) else self.path(dst),
        )

    def _upload(self, src, dst):
        if src.is_dir():
            if not dst.exists():
                self.sftp.mkdir(str(dst))
            for fn in src:
                self._upload(fn, dst / fn.name)
        elif dst.is_dir():
            self.sftp.put(str(src), str(dst / src.name))
        else:
            self.sftp.put(str(src), str(dst))

    def connect_sock(self, dport, dhost="localhost", ipv6=False):
        """Returns a Paramiko ``Channel``, connected to dhost:dport on the remote machine.
        The ``Channel`` behaves like a regular socket; you can ``send`` and ``recv`` on it
        and the data will pass encrypted over SSH. Usage::

            mach = ParamikoMachine("myhost")
            sock = mach.connect_sock(12345)
            data = sock.recv(100)
            sock.send("foobar")
            sock.close()
        """
        if ipv6 and dhost == "localhost":
            dhost = "::1"
        srcaddr = ("::1", 0, 0, 0) if ipv6 else ("127.0.0.1", 0)
        trans = self._client.get_transport()
        trans.set_keepalive(self._keep_alive)
        chan = trans.open_channel("direct-tcpip", (dhost, dport), srcaddr)
        return SocketCompatibleChannel(chan)

    #
    # Path implementation
    #
    def _path_listdir(self, fn):
        return self.sftp.listdir(str(fn))

    def _path_read(self, fn):
        f = self.sftp.open(str(fn), "rb")
        data = f.read()
        f.close()
        return data

    def _path_write(self, fn, data):
        if self.custom_encoding and isinstance(data, str):
            data = data.encode(self.custom_encoding)
        f = self.sftp.open(str(fn), "wb")
        f.write(data)
        f.close()

    def _path_stat(self, fn):
        try:
            st = self.sftp.stat(str(fn))
        except OSError as e:
            if e.errno == errno.ENOENT:
                return None
            raise
        res = StatRes(
            (
                st.st_mode,
                0,
                0,
                0,
                st.st_uid,
                st.st_gid,
                st.st_size,
                st.st_atime,
                st.st_mtime,
                0,
            )
        )

        if stat.S_ISDIR(st.st_mode):
            res.text_mode = "directory"
        if stat.S_ISREG(st.st_mode):
            res.text_mode = "regular file"
        return res

    def daemonic_popen(self, command, cwd="/", stdout=None, stderr=None, append=True):
        raise NotImplementedError("This is not implemented on ParamikoMachine!")


###################################################################################################
# Make paramiko.Channel adhere to the socket protocol, namely, send and recv should fail
# when the socket has been closed
###################################################################################################
class SocketCompatibleChannel:
    def __init__(self, chan):
        self._chan = chan

    def __getattr__(self, name):
        return getattr(self._chan, name)

    def send(self, s):
        if self._chan.closed:
            raise OSError(errno.EBADF, "Bad file descriptor")
        return self._chan.send(s)

    def recv(self, count):
        if self._chan.closed:
            raise OSError(errno.EBADF, "Bad file descriptor")
        return self._chan.recv(count)


###################################################################################################
# Custom iter_lines for paramiko.Channel
###################################################################################################
def _iter_lines(
    proc,
    decode,  # noqa: ARG001
    linesize,
    line_timeout=None,
):
    from selectors import EVENT_READ, DefaultSelector

    # Python 3.4+ implementation
    def selector():
        sel = DefaultSelector()
        sel.register(proc.stdout.channel, EVENT_READ)
        while True:
            ready = sel.select(line_timeout)
            if not ready and line_timeout:
                raise ProcessLineTimedOut(
                    "popen line timeout expired",
                    getattr(proc, "argv", None),
                    getattr(proc, "machine", None),
                )
            for _key, _mask in ready:
                yield

    for _ in selector():
        if proc.stdout.channel.recv_ready():
            yield 0, proc.stdout.readline(linesize)
        if proc.stdout.channel.recv_stderr_ready():
            yield 1, proc.stderr.readline(linesize)
        if proc.poll() is not None:
            break

    for line in proc.stdout:
        yield 0, line
    for line in proc.stderr:
        yield 1, line



from __future__ import annotations

import os
import platform
import sys

import pytest

skip_without_chown = pytest.mark.skipif(
    not hasattr(os, "chown"), reason="os.chown not supported"
)

skip_without_tty = pytest.mark.skipif(not sys.stdin.isatty(), reason="Not a TTY")

skip_on_windows = pytest.mark.skipif(
    sys.platform == "win32", reason="Windows not supported for this test (yet)"
)

xfail_on_windows = pytest.mark.xfail(
    sys.platform == "win32", reason="Windows not supported for this test (yet)"
)

xfail_on_pypy = pytest.mark.xfail(
    platform.python_implementation() == "PyPy",
    reason="PyPy is currently not working on this test!",
)



from __future__ import annotations

import contextlib
import re
from tempfile import NamedTemporaryFile

from plumbum.commands import CommandNotFound, ConcreteCommand, shquote
from plumbum.lib import ProcInfo
from plumbum.machines.base import BaseMachine
from plumbum.machines.env import BaseEnv
from plumbum.path.local import LocalPath
from plumbum.path.remote import RemotePath, RemoteWorkdir, StatRes


class RemoteEnv(BaseEnv):
    """The remote machine's environment; exposes a dict-like interface"""

    __slots__ = ["_orig", "remote"]

    def __init__(self, remote):
        session = remote._session
        # GNU env has a -0 argument; use it if present. Otherwise,
        # fall back to calling printenv on each (possible) variable
        # from plain env.
        env0 = session.run("env -0; echo")
        if env0[0] == 0 and not env0[2].rstrip():
            _curr = dict(
                line.split("=", 1) for line in env0[1].split("\x00") if "=" in line
            )
        else:
            lines = session.run("env; echo")[1].splitlines()
            split = (line.split("=", 1) for line in lines)
            keys = (line[0] for line in split if len(line) > 1)
            runs = ((key, session.run(f'printenv "{key}"; echo')) for key in keys)
            _curr = {
                key: run[1].rstrip("\n")
                for (key, run) in runs
                if run[0] == 0 and run[1].rstrip("\n") and not run[2]
            }

        super().__init__(remote.path, ":", _curr=_curr)
        self.remote = remote
        self._orig = self._curr.copy()

    def __delitem__(self, name):
        BaseEnv.__delitem__(self, name)
        self.remote._session.run(f"unset {name}")

    def __setitem__(self, name, value):
        BaseEnv.__setitem__(self, name, value)
        self.remote._session.run(f"export {name}={shquote(value)}")

    def pop(self, name, *default):
        BaseEnv.pop(self, name, *default)
        self.remote._session.run(f"unset {name}")

    def update(self, *args, **kwargs):
        BaseEnv.update(self, *args, **kwargs)
        self.remote._session.run(
            "export " + " ".join(f"{k}={shquote(v)}" for k, v in self.getdict().items())
        )

    def expand(self, expr):
        """Expands any environment variables and home shortcuts found in ``expr``
        (like ``os.path.expanduser`` combined with ``os.path.expandvars``)

        :param expr: An expression containing environment variables (as ``$FOO``) or
                     home shortcuts (as ``~/.bashrc``)

        :returns: The expanded string"""
        return self.remote.expand(expr)

    def expanduser(self, expr):
        """Expand home shortcuts (e.g., ``~/foo/bar`` or ``~john/foo/bar``)

        :param expr: An expression containing home shortcuts

        :returns: The expanded string"""
        return self.remote.expanduser(expr)

    # def clear(self):
    #    BaseEnv.clear(self, *args, **kwargs)
    #    self.remote._session.run("export %s" % " ".join("%s=%s" % (k, v) for k, v in self.getdict()))

    def getdelta(self):
        """Returns the difference between the this environment and the original environment of
        the remote machine"""
        self._curr["PATH"] = self.path.join()

        delta = {}
        for k, v in self._curr.items():
            if k not in self._orig:
                delta[k] = str(v)
        for k, v in self._orig.items():
            if k not in self._curr:
                delta[k] = ""
            else:
                if v != self._curr[k]:
                    delta[k] = self._curr[k]

        return delta


class RemoteCommand(ConcreteCommand):
    __slots__ = ("remote",)
    QUOTE_LEVEL = 1

    def __init__(self, remote, executable, encoding="auto"):
        self.remote = remote
        ConcreteCommand.__init__(
            self, executable, remote.custom_encoding if encoding == "auto" else encoding
        )

    @property
    def machine(self):
        return self.remote

    def __repr__(self):
        return f"RemoteCommand({self.remote!r}, {self.executable!r})"

    def popen(self, args=(), **kwargs):
        return self.remote.popen(self[args], **kwargs)

    def nohup(self, cwd=".", stdout="nohup.out", stderr=None, append=True):
        """Runs a command detached."""
        return self.machine.daemonic_popen(self, cwd, stdout, stderr, append)


class ClosedRemoteMachine(Exception):
    pass


class ClosedRemote:
    __slots__ = ["__weakref__", "_obj"]

    def __init__(self, obj):
        self._obj = obj

    def close(self):
        pass

    def __getattr__(self, name):
        raise ClosedRemoteMachine(f"{self._obj!r} has been closed")


class BaseRemoteMachine(BaseMachine):
    """Represents a *remote machine*; serves as an entry point to everything related to that
    remote machine, such as working directory and environment manipulation, command creation,
    etc.

    Attributes:

    * ``cwd`` - the remote working directory
    * ``env`` - the remote environment
    * ``custom_encoding`` - the remote machine's default encoding (assumed to be UTF8)
    * ``connect_timeout`` - the connection timeout


    There also is a _cwd attribute that exists if the cwd is not current (del if cwd is changed).
    """

    # allow inheritors to override the RemoteCommand class
    RemoteCommand = RemoteCommand

    @property
    def cwd(self):
        if not hasattr(self, "_cwd"):
            self._cwd = RemoteWorkdir(self)
        return self._cwd

    def __init__(self, encoding="utf8", connect_timeout=10, new_session=False):
        self.custom_encoding = encoding
        self.connect_timeout = connect_timeout
        self._session = self.session(new_session=new_session)
        self.uname = self._get_uname()
        self.env = RemoteEnv(self)
        self._python = None
        self._program_cache = {}

    def _get_uname(self):
        rc, out, _ = self._session.run("uname", retcode=None)
        if rc == 0:
            return out.strip()

        rc, out, _ = self._session.run(
            "python3 -c 'import platform;print(platform.uname()[0])'", retcode=None
        )
        if rc == 0:
            return out.strip()

        # all POSIX systems should have uname. make an educated guess it's Windows
        return "Windows"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()

    def close(self):
        """closes the connection to the remote machine; all paths and programs will
        become defunct"""
        self._session.close()
        self._session = ClosedRemote(self)

    def path(self, *parts):
        """A factory for :class:`RemotePaths <plumbum.path.remote.RemotePath>`.
        Usage: ``p = rem.path("/usr", "lib", "python2.7")``
        """
        parts2 = [str(self.cwd)]
        for p in parts:
            if isinstance(p, LocalPath):
                raise TypeError(f"Cannot construct RemotePath from {p!r}")
            parts2.append(self.expanduser(str(p)))
        return RemotePath(self, *parts2)

    def which(self, progname):
        """Looks up a program in the ``PATH``. If the program is not found, raises
        :class:`CommandNotFound <plumbum.commands.CommandNotFound>`

        :param progname: The program's name. Note that if underscores (``_``) are present
                         in the name, and the exact name is not found, they will be replaced
                         in turn by hyphens (``-``) then periods (``.``), and the name will
                         be looked up again for each alternative

        :returns: A :class:`RemotePath <plumbum.path.local.RemotePath>`
        """
        key = (progname, self.env.get("PATH", ""))

        with contextlib.suppress(KeyError):
            return self._program_cache[key]

        alternatives = [progname]
        if "_" in progname:
            alternatives += [progname.replace("_", "-"), progname.replace("_", ".")]
        for name in alternatives:
            for p in self.env.path:
                fn = p / name
                if fn.access("x") and not fn.is_dir():
                    self._program_cache[key] = fn
                    return fn

        raise CommandNotFound(progname, self.env.path)

    def __getitem__(self, cmd):
        """Returns a `Command` object representing the given program. ``cmd`` can be a string or
        a :class:`RemotePath <plumbum.path.remote.RemotePath>`; if it is a path, a command
        representing this path will be returned; otherwise, the program name will be looked up in
        the system's ``PATH`` (using ``which``). Usage::

            r_ls = rem["ls"]
        """
        if isinstance(cmd, RemotePath):
            if cmd.remote is self:
                return self.RemoteCommand(self, cmd)

            raise TypeError(
                f"Given path does not belong to this remote machine: {cmd!r}"
            )

        if not isinstance(cmd, LocalPath):
            return self.RemoteCommand(
                self, self.path(cmd) if "/" in cmd or "\\" in cmd else self.which(cmd)
            )

        raise TypeError(f"cmd must not be a LocalPath: {cmd!r}")

    @property
    def python(self):
        """A command that represents the default remote python interpreter"""
        if not self._python:
            self._python = self["python3"]
        return self._python

    def session(self, isatty=False, *, new_session=False):
        """Creates a new :class:`ShellSession <plumbum.session.ShellSession>` object; this invokes the user's
        shell on the remote machine and executes commands on it over stdin/stdout/stderr
        """
        raise NotImplementedError()

    def download(self, src, dst):
        """Downloads a remote file/directory (``src``) to a local destination (``dst``).
        ``src`` must be a string or a :class:`RemotePath <plumbum.path.remote.RemotePath>`
        pointing to this remote machine, and ``dst`` must be a string or a
        :class:`LocalPath <plumbum.machines.local.LocalPath>`"""
        raise NotImplementedError()

    def upload(self, src, dst):
        """Uploads a local file/directory (``src``) to a remote destination (``dst``).
        ``src`` must be a string or a :class:`LocalPath <plumbum.machines.local.LocalPath>`,
        and ``dst`` must be a string or a :class:`RemotePath <plumbum.path.remote.RemotePath>`
        pointing to this remote machine"""
        raise NotImplementedError()

    def popen(self, args, **kwargs):
        """Spawns the given command on the remote machine, returning a ``Popen``-like object;
        do not use this method directly, unless you need "low-level" control on the remote
        process"""
        raise NotImplementedError()

    def list_processes(self):
        """
        Returns information about all running processes (on POSIX systems: using ``ps``)

        .. versionadded:: 1.3
        """
        ps = self["ps"]
        lines = ps("-e", "-o", "pid,uid,stat,args").splitlines()
        lines.pop(0)  # header
        for line in lines:
            parts = line.strip().split()
            yield ProcInfo(int(parts[0]), int(parts[1]), parts[2], " ".join(parts[3:]))

    def pgrep(self, pattern):
        """
        Process grep: return information about all processes whose command-line args match the given regex pattern
        """
        pat = re.compile(pattern)
        for procinfo in self.list_processes():
            if pat.search(procinfo.args):
                yield procinfo

    @contextlib.contextmanager
    def tempdir(self):
        """A context manager that creates a remote temporary directory, which is removed when
        the context exits"""
        _, out, _ = self._session.run(
            "mktemp -d 2>/dev/null || mktemp -d tmp.XXXXXXXXXX"
        )
        local_dir = self.path(out.strip())
        try:
            yield local_dir
        finally:
            local_dir.delete()

    #
    # Path implementation
    #
    def _path_listdir(self, fn):
        files = self._session.run(f"ls -a {shquote(fn)}")[1].splitlines()
        files.remove(".")
        files.remove("..")
        return files

    def _path_glob(self, fn, pattern):
        # shquote does not work here due to the way bash loops use space as a separator
        pattern = pattern.replace(" ", r"\ ")
        fn = fn.replace(" ", r"\ ")
        matches = self._session.run(rf"for fn in {fn}/{pattern}; do echo $fn; done")[
            1
        ].splitlines()
        if len(matches) == 1 and not self._path_stat(matches[0]):
            return []  # pattern expansion failed
        return matches

    def _path_getuid(self, fn):
        stat_cmd = (
            "stat -c '%u,%U' "
            if self.uname not in ("Darwin", "FreeBSD")
            else "stat -f '%u,%Su' "
        )
        return self._session.run(stat_cmd + shquote(fn))[1].strip().split(",")

    def _path_getgid(self, fn):
        stat_cmd = (
            "stat -c '%g,%G' "
            if self.uname not in ("Darwin", "FreeBSD")
            else "stat -f '%g,%Sg' "
        )
        return self._session.run(stat_cmd + shquote(fn))[1].strip().split(",")

    def _path_stat(self, fn):
        if self.uname not in ("Darwin", "FreeBSD"):
            stat_cmd = "stat -c '%F,%f,%i,%d,%h,%u,%g,%s,%X,%Y,%Z' "
        else:
            stat_cmd = "stat -f '%HT,%Xp,%i,%d,%l,%u,%g,%z,%a,%m,%c' "
        rc, out, _ = self._session.run(stat_cmd + shquote(fn), retcode=None)
        if rc != 0:
            return None
        statres = out.strip().split(",")
        text_mode = statres.pop(0).lower()
        res = StatRes((int(statres[0], 16), *tuple(int(sr) for sr in statres[1:])))
        res.text_mode = text_mode
        return res

    def _path_delete(self, fn):
        self._session.run(f"rm -rf {shquote(fn)}")

    def _path_move(self, src, dst):
        self._session.run(f"mv {shquote(src)} {shquote(dst)}")

    def _path_copy(self, src, dst):
        self._session.run(f"cp -r {shquote(src)} {shquote(dst)}")

    def _path_mkdir(
        self,
        fn,
        mode=None,  # noqa: ARG002
        minus_p=True,
    ):
        p_str = "-p " if minus_p else ""
        cmd = f"mkdir {p_str}{shquote(fn)}"
        self._session.run(cmd)

    def _path_chmod(self, mode, fn):
        self._session.run(f"chmod {mode:o} {shquote(fn)}")

    def _path_touch(self, path):
        self._session.run(f"touch {path}")

    def _path_chown(self, fn, owner, group, recursive):
        args = ["chown"]
        if recursive:
            args.append("-R")
        if owner is not None and group is not None:
            args.append(f"{owner}:{group}")
        elif owner is not None:
            args.append(str(owner))
        elif group is not None:
            args.append(f":{group}")
        args.append(shquote(fn))
        self._session.run(" ".join(args))

    def _path_read(self, fn):
        data = self["cat"](fn)
        if self.custom_encoding and isinstance(data, str):
            return data.encode(self.custom_encoding)
        return data

    def _path_write(self, fn, data):
        if self.custom_encoding and isinstance(data, str):
            data = data.encode(self.custom_encoding)
        with NamedTemporaryFile() as f:
            f.write(data)
            f.flush()
            f.seek(0)
            self.upload(f.name, fn)

    def _path_link(self, src, dst, symlink):
        symlink_str = "-s " if symlink else ""
        self._session.run(f"ln {symlink_str}{shquote(src)} {shquote(dst)}")

    def expand(self, expr):
        return self._session.run(f"echo {expr}")[1].strip()

    def expanduser(self, expr):
        if not any(part.startswith("~") for part in expr.split("/")):
            return expr
        # we escape all $ signs to avoid expanding env-vars
        expr_repl = expr.replace("$", "\\$")
        return self._session.run(f"echo {expr_repl}")[1].strip()



"""
Names for the standard and extended color set.
Extended set is similar to `vim wiki <http://vim.wikia.com/wiki/Xterm256_color_names_for_console_Vim>`_, `colored <https://pypi.python.org/pypi/colored>`_, etc. Colors based on `wikipedia <https://en.wikipedia.org/wiki/ANSI_escape_code#Colors>`_.

You can access the index of the colors with names.index(name). You can access the
rgb values with ``r=int(html[n][1:3],16)``, etc.
"""

from __future__ import annotations

color_names = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_gray",
    "dark_gray",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
    "grey_0",
    "navy_blue",
    "dark_blue",
    "blue_3",
    "blue_3a",
    "blue_1",
    "dark_green",
    "deep_sky_blue_4",
    "deep_sky_blue_4a",
    "deep_sky_blue_4b",
    "dodger_blue_3",
    "dodger_blue_2",
    "green_4",
    "spring_green_4",
    "turquoise_4",
    "deep_sky_blue_3",
    "deep_sky_blue_3a",
    "dodger_blue_1",
    "green_3",
    "spring_green_3",
    "dark_cyan",
    "light_sea_green",
    "deep_sky_blue_2",
    "deep_sky_blue_1",
    "green_3a",
    "spring_green_3a",
    "spring_green_2",
    "cyan_3",
    "dark_turquoise",
    "turquoise_2",
    "green_1",
    "spring_green_2a",
    "spring_green_1",
    "medium_spring_green",
    "cyan_2",
    "cyan_1",
    "dark_red",
    "deep_pink_4",
    "purple_4",
    "purple_4a",
    "purple_3",
    "blue_violet",
    "orange_4",
    "grey_37",
    "medium_purple_4",
    "slate_blue_3",
    "slate_blue_3a",
    "royal_blue_1",
    "chartreuse_4",
    "dark_sea_green_4",
    "pale_turquoise_4",
    "steel_blue",
    "steel_blue_3",
    "cornflower_blue",
    "chartreuse_3",
    "dark_sea_green_4a",
    "cadet_blue",
    "cadet_blue_a",
    "sky_blue_3",
    "steel_blue_1",
    "chartreuse_3a",
    "pale_green_3",
    "sea_green_3",
    "aquamarine_3",
    "medium_turquoise",
    "steel_blue_1a",
    "chartreuse_2a",
    "sea_green_2",
    "sea_green_1",
    "sea_green_1a",
    "aquamarine_1",
    "dark_slate_gray_2",
    "dark_red_a",
    "deep_pink_4a",
    "dark_magenta",
    "dark_magenta_a",
    "dark_violet",
    "purple",
    "orange_4a",
    "light_pink_4",
    "plum_4",
    "medium_purple_3",
    "medium_purple_3a",
    "slate_blue_1",
    "yellow_4",
    "wheat_4",
    "grey_53",
    "light_slate_grey",
    "medium_purple",
    "light_slate_blue",
    "yellow_4_a",
    "dark_olive_green_3",
    "dark_sea_green",
    "light_sky_blue_3",
    "light_sky_blue_3a",
    "sky_blue_2",
    "chartreuse_2",
    "dark_olive_green_3a",
    "pale_green_3a",
    "dark_sea_green_3",
    "dark_slate_gray_3",
    "sky_blue_1",
    "chartreuse_1",
    "light_green_a",
    "light_green_b",
    "pale_green_1",
    "aquamarine_1a",
    "dark_slate_gray_1",
    "red_3",
    "deep_pink_4b",
    "medium_violet_red",
    "magenta_3",
    "dark_violet_a",
    "purple_a",
    "dark_orange_3",
    "indian_red",
    "hot_pink_3",
    "medium_orchid_3",
    "medium_orchid",
    "medium_purple_2",
    "dark_goldenrod",
    "light_salmon_3",
    "rosy_brown",
    "grey_63",
    "medium_purple_2a",
    "medium_purple_1",
    "gold_3",
    "dark_khaki",
    "navajo_white_3",
    "grey_69",
    "light_steel_blue_3",
    "light_steel_blue",
    "yellow_3",
    "dark_olive_green_3b",
    "dark_sea_green_3a",
    "dark_sea_green_2",
    "light_cyan_3",
    "light_sky_blue_1",
    "green_yellow",
    "dark_olive_green_2",
    "pale_green_1a",
    "dark_sea_green_2a",
    "dark_sea_green_1",
    "pale_turquoise_1",
    "red_3a",
    "deep_pink_3",
    "deep_pink_3a",
    "magenta_3a",
    "magenta_3b",
    "magenta_2",
    "dark_orange_3a",
    "indian_red_a",
    "hot_pink_3a",
    "hot_pink_2",
    "orchid",
    "medium_orchid_1",
    "orange_3",
    "light_salmon_3a",
    "light_pink_3",
    "pink_3",
    "plum_3",
    "violet",
    "gold_3a",
    "light_goldenrod_3",
    "tan",
    "misty_rose_3",
    "thistle_3",
    "plum_2",
    "yellow_3a",
    "khaki_3",
    "light_goldenrod_2",
    "light_yellow_3",
    "grey_84",
    "light_steel_blue_1",
    "yellow_2",
    "dark_olive_green_1",
    "dark_olive_green_1a",
    "dark_sea_green_1a",
    "honeydew_2",
    "light_cyan_1",
    "red_1",
    "deep_pink_2",
    "deep_pink_1",
    "deep_pink_1a",
    "magenta_2a",
    "magenta_1",
    "orange_red_1",
    "indian_red_1",
    "indian_red_1a",
    "hot_pink",
    "hot_pink_a",
    "medium_orchid_1a",
    "dark_orange",
    "salmon_1",
    "light_coral",
    "pale_violet_red_1",
    "orchid_2",
    "orchid_1",
    "orange_1",
    "sandy_brown",
    "light_salmon_1",
    "light_pink_1",
    "pink_1",
    "plum_1",
    "gold_1",
    "light_goldenrod_2a",
    "light_goldenrod_2b",
    "navajo_white_1",
    "misty_rose_1",
    "thistle_1",
    "yellow_1",
    "light_goldenrod_1",
    "khaki_1",
    "wheat_1",
    "cornsilk_1",
    "grey_10_0",
    "grey_3",
    "grey_7",
    "grey_11",
    "grey_15",
    "grey_19",
    "grey_23",
    "grey_27",
    "grey_30",
    "grey_35",
    "grey_39",
    "grey_42",
    "grey_46",
    "grey_50",
    "grey_54",
    "grey_58",
    "grey_62",
    "grey_66",
    "grey_70",
    "grey_74",
    "grey_78",
    "grey_82",
    "grey_85",
    "grey_89",
    "grey_93",
]

EMPTY_SLICE = slice(None, None, None)

_greys = (
    3.4,
    7.4,
    11,
    15,
    19,
    23,
    26.7,
    30.49,
    34.6,
    38.6,
    42.4,
    46.4,
    50,
    54,
    58,
    62,
    66,
    69.8,
    73.8,
    77.7,
    81.6,
    85.3,
    89.3,
    93,
)
_grey_vals = [int(x / 100.0 * 16 * 16) for x in _greys]

_grey_html = ["#" + format(x, "02x") * 3 for x in _grey_vals]

_normals = [int(x, 16) for x in ["0", "5f", "87", "af", "d7", "ff"]]
_normal_html = [
    "#"
    + format(_normals[n // 36], "02x")
    + format(_normals[n // 6 % 6], "02x")
    + format(_normals[n % 6], "02x")
    for n in range(16 - 16, 232 - 16)
]

_base_pattern = [(n // 4, n // 2 % 2, n % 2) for n in range(8)]
_base_html = (
    [f"#{x[2] * 192:02x}{x[1] * 192:02x}{x[0] * 192:02x}" for x in _base_pattern]
    + ["#808080"]
    + [f"#{x[2] * 255:02x}{x[1] * 255:02x}{x[0] * 255:02x}" for x in _base_pattern][1:]
)
color_html = _base_html + _normal_html + _grey_html

color_codes_simple = list(range(8)) + list(range(60, 68))
"""Simple colors, remember that reset is #9, second half is non as common."""

# Attributes
attributes_ansi = {
    "bold": 1,
    "dim": 2,
    "italics": 3,
    "underline": 4,
    "reverse": 7,
    "hidden": 8,
    "strikeout": 9,
}

# Stylesheet
default_styles = {
    "warn": "fg red",
    "title": "fg cyan underline bold",
    "fatal": "fg red bold",
    "highlight": "bg yellow",
    "info": "fg blue",
    "success": "fg green",
}

# Functions to be used for color name operations


class FindNearest:
    """This is a class for finding the nearest color given rgb values.
    Different find methods are available."""

    def __init__(self, r: int, g: int, b: int) -> None:
        self.r = r
        self.b = b
        self.g = g

    def only_basic(self):
        """This will only return the first 8 colors!
        Breaks the colorspace into cubes, returns color"""
        midlevel = 0x40  # Since bright is not included

        # The colors are organised so that it is a
        # 3D cube, black at 0,0,0, white at 1,1,1
        # Compressed to linear_integers r,g,b
        # [[[0,1],[2,3]],[[4,5],[6,7]]]
        # r*1 + g*2 + b*4
        return (
            (self.r >= midlevel) * 1
            + (self.g >= midlevel) * 2
            + (self.b >= midlevel) * 4
        )

    def all_slow(self, color_slice: slice = EMPTY_SLICE) -> int:
        """This is a slow way to find the nearest color."""
        distances = [
            self._distance_to_color(color) for color in color_html[color_slice]
        ]
        return min(range(len(distances)), key=distances.__getitem__)

    def _distance_to_color(self, color: str) -> int:
        """This computes the distance to a color, should be minimized."""
        rgb = (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16))
        return (self.r - rgb[0]) ** 2 + (self.g - rgb[1]) ** 2 + (self.b - rgb[2]) ** 2

    def _distance_to_color_number(self, n: int) -> int:
        color = color_html[n]
        return self._distance_to_color(color)

    def only_colorblock(self) -> int:
        """This finds the nearest color based on block system, only works
        for 17-232 color values."""
        rint = min(
            range(len(_normals)), key=[abs(x - self.r) for x in _normals].__getitem__
        )
        bint = min(
            range(len(_normals)), key=[abs(x - self.b) for x in _normals].__getitem__
        )
        gint = min(
            range(len(_normals)), key=[abs(x - self.g) for x in _normals].__getitem__
        )
        return 16 + 36 * rint + 6 * gint + bint

    def only_simple(self) -> int:
        """Finds the simple color-block color."""
        return self.all_slow(slice(0, 16, None))

    def only_grey(self) -> int:
        """Finds the greyscale color."""
        rawval = (self.r + self.b + self.g) / 3
        n = min(
            range(len(_grey_vals)),
            key=[abs(x - rawval) for x in _grey_vals].__getitem__,
        )
        return n + 232

    def all_fast(self) -> int:
        """Runs roughly 8 times faster than the slow version."""
        colors = [self.only_simple(), self.only_colorblock(), self.only_grey()]
        distances = [self._distance_to_color_number(n) for n in colors]
        return colors[min(range(len(distances)), key=distances.__getitem__)]


def from_html(color: str) -> tuple[int, int, int]:
    """Convert html hex code to rgb."""
    if len(color) != 7 or color[0] != "#":
        raise ValueError("Invalid length of html code")
    return (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16))


def to_html(r, g, b):
    """Convert rgb to html hex code."""
    return f"#{r:02x}{g:02x}{b:02x}"



#!/usr/bin/env python3
from __future__ import annotations

from plumbum import colors

with colors:
    print("Do you believe in color, punk? DO YOU?")
    for i in range(0, 255, 10):
        for j in range(0, 255, 10):
            print("".join(colors.rgb(i, j, k)["\u2588"] for k in range(0, 255, 10)))



from __future__ import annotations

from plumbum import cli, colors

# from plumbum.colorlib import HTMLStyle, StyleFactory
# plumbum.colors = StyleFactory(HTMLStyle)


class MyApp(cli.Application):
    PROGNAME = colors.green
    VERSION = colors.blue | "1.0.2"
    COLOR_GROUPS = {"Meta-switches": colors.yellow}
    COLOR_GROUP_TITLES = {"Meta-switches": colors.bold & colors.yellow}
    opts = cli.Flag("--ops", help=colors.magenta | "This is help")

    def main(self):
        print("HI")


if __name__ == "__main__":
    MyApp.run()



from __future__ import annotations

from plumbum.path.base import FSUser, Path, RelativePath
from plumbum.path.local import LocalPath, LocalWorkdir
from plumbum.path.remote import RemotePath, RemoteWorkdir
from plumbum.path.utils import copy, delete, move

__all__ = (
    "FSUser",
    "LocalPath",
    "LocalWorkdir",
    "Path",
    "RelativePath",
    "RemotePath",
    "RemoteWorkdir",
    "copy",
    "delete",
    "move",
)



from __future__ import annotations

import inspect
import os
import sys
from contextlib import contextmanager
from io import StringIO

IS_WIN32 = sys.platform == "win32"


class ProcInfo:
    def __init__(self, pid, uid, stat, args):
        self.pid = pid
        self.uid = uid
        self.stat = stat
        self.args = args

    def __repr__(self):
        return f"ProcInfo({self.pid!r}, {self.uid!r}, {self.stat!r}, {self.args!r})"


@contextmanager
def captured_stdout(stdin=""):
    """
    Captures stdout (similar to the redirect_stdout in Python 3.4+, but with slightly different arguments)
    """
    prevstdin = sys.stdin
    prevstdout = sys.stdout
    sys.stdin = StringIO(stdin)
    sys.stdout = StringIO()
    try:
        yield sys.stdout
    finally:
        sys.stdin = prevstdin
        sys.stdout = prevstdout


class StaticProperty:
    """This acts like a static property, allowing access via class or object.
    This is a non-data descriptor."""

    def __init__(self, function):
        self._function = function
        self.__doc__ = function.__doc__

    def __get__(self, obj, klass=None):
        return self._function()


def getdoc(obj):
    """
    This gets a docstring if available, and cleans it, but does not look up docs in
    inheritance tree (Pre Python 3.5 behavior of ``inspect.getdoc``).
    """
    try:
        doc = obj.__doc__
    except AttributeError:
        return None
    if not isinstance(doc, str):
        return None
    return inspect.cleandoc(doc)


def read_fd_decode_safely(fd, size=4096):
    """
    This reads a utf-8 file descriptor and returns a chunk, growing up to
    three bytes if needed to decode the character at the end.

    Returns the data and the decoded text.
    """
    data = os.read(fd.fileno(), size)
    for _ in range(3):
        try:
            return data, data.decode("utf-8")
        except UnicodeDecodeError as e:
            if e.reason != "unexpected end of data":
                raise
            data += os.read(fd.fileno(), 1)

    return data, data.decode("utf-8")



#!/usr/bin/env python3
from __future__ import annotations

import os
import unittest

from plumbum import colors

# This is really intended to be run manually, so the output can be observed, rather than with py.test


class TestVisualColor(unittest.TestCase):
    def setUp(self):
        if os.name == "nt":
            try:
                import colorama

                colorama.init()
                self.colorama = colorama
                colors.use_color = 1
                print()
                print("Colorama initialized")
            except ImportError:
                self.colorama = None
        else:
            self.colorama = None

    def tearDown(self):
        if self.colorama:
            self.colorama.deinit()

    def testVisualColors(self):
        print()
        for c in colors.fg[:16]:
            with c:
                print("Cycle color test", end=" ")
            print(" - > back to normal")
        with colors:
            print(
                colors.fg.green
                + "Green "
                + colors.bold
                + "Bold "
                + ~colors.bold
                + "Normal"
            )
        print("Reset all")

    def testToggleColors(self):
        print()
        print(colors.fg.red["This is in red"], "but this is not")
        print(
            colors.fg.green
            + "Hi, "
            + colors.bg[23]
            + "This is on a BG"
            + ~colors.bg
            + " and this is not but is still green."
        )
        colors.yellow.print("This is printed from color.")
        colors.reset()

        for attr in colors._style.attribute_names:
            print("This is", attr | getattr(colors, attr), "and this is not.")
            colors.reset()

    def testLimits(self):
        print()
        cval = colors.use_color
        colors.use_color = 4
        c = colors.rgb(123, 40, 200)
        print("True", repr(str(c)), repr(c))
        colors.use_color = 3
        print("Full", repr(str(c)), repr(c))
        colors.use_color = 2
        print("Simple", repr(str(c)), repr(c))
        colors.use_color = 1
        print("Basic", repr(str(c)), repr(c))
        colors.use_color = 0
        print("None", repr(str(c)), repr(c))
        colors.use_color = cval


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from .application import Application
from .config import Config, ConfigINI
from .switches import (
    CSV,
    CountOf,
    ExistingDirectory,
    ExistingFile,
    Flag,
    NonexistentPath,
    Predicate,
    Range,
    Set,
    SwitchAttr,
    SwitchError,
    autoswitch,
    positional,
    switch,
)

__all__ = (
    "CSV",
    "Application",
    "Config",
    "ConfigINI",
    "CountOf",
    "ExistingDirectory",
    "ExistingFile",
    "Flag",
    "NonexistentPath",
    "Predicate",
    "Range",
    "Set",
    "SwitchAttr",
    "SwitchError",
    "autoswitch",
    "positional",
    "switch",
)



from __future__ import annotations

from plumbum import cli


class Main3Validator(cli.Application):
    def main(self, myint: int, myint2: int, *mylist: int):
        print(myint, myint2, mylist)


class TestProg3:
    def test_prog(self, capsys):
        _, rc = Main3Validator.run(["prog", "1", "2", "3", "4", "5"], exit=False)
        assert rc == 0
        assert "1 2 (3, 4, 5)" in capsys.readouterr()[0]


class Main4Validator(cli.Application):
    def main(self, myint: int, myint2: int, *mylist: int) -> None:
        print(myint, myint2, mylist)


class TestProg4:
    def test_prog(self, capsys):
        _, rc = Main4Validator.run(["prog", "1", "2", "3", "4", "5"], exit=False)
        assert rc == 0
        assert "1 2 (3, 4, 5)" in capsys.readouterr()[0]



from __future__ import annotations

import pytest

from plumbum import local
from plumbum._testtools import skip_on_windows

pytestmark = pytest.mark.sudo

# This is a separate file to make separating (ugly) sudo command easier
# For example, you can now run test_local directly without typing a password


class TestSudo:
    @skip_on_windows
    def test_as_user(self):
        with local.as_root():
            local["date"]()
